// ----------------------------------------------------------------------------------------------------------------------
//   Copyright 2014 (c) Microsoft Corporation.
//   Summary: Defines the Graph.
// ----------------------------------------------------------------------------------------------------------------------

//
//  Prototype hierarchy
//      Terms
//      Term
//          Boolean
//          Integer 
//          Double 
//          RDFDate
//          GeometryPoint
//          Text 
//          Variable
//          Symbol
//          Entity
//              Person
//              Preferences
//              Favorites
//              PostalAddress
//              GeoAnnotation
//                  Place
//                  Route
//          
//  Object hierarchy
//      Graph
//          Me (person)
//              Preferences
//                  Favorites

(function () {
    "use strict"; // whole file
    
    //      ___                            _
    //     |_ _|_ __ ___  _ __   ___  _ __| |_ ___
    //      | || '_ ` _ \| '_ \ / _ \| '__| __/ __|
    //      | || | | | | | |_) | (_) | |  | |_\__ \
    //     |___|_| |_| |_| .__/ \___/|_|   \__|___/
    //                   |_|
    var utilities = require("./Utilities");
    var config = require("./Configuration").Configuration;
    var log = require("./Logger");
    var enums = require("./Enum");
    var tokenizer = require("./Tokenizer");
    
    //      ____                 _
    //     / ___|_ __ __ _ _ __ | |__
    //    | |  _| '__/ _` | '_ \| '_ \
    //    | |_| | | | (_| | |_) | | | |
    //     \____|_|  \__,_| .__/|_| |_|
    //                    |_|
    function Graph(callback, debug) {
        this._extends("_etag", { value: "0", writable: true});
        this._extends("_mapCache", { value: {} });
        this._extends("_recordDelta", { value: true, writable: true });
        
        this._extends("isNodeJS", { value: typeof (process) !== "undefined" });
        
        // http
        //  Calls the web server from both nodejs and the browser.
        var http = function (url, method, headers, requestBody, callback) {
            var engine = this;
            if (this.isNodeJS) {
                var options = require("url").parse(url);
                options.method = method;
                options.headers = headers;
                
                // TODO: Remove workaround for fiddler
                if (debug) {
                    options.pathname = options.path = url;
                    options.headers.host = options.host;
                    options.hostname = options.host = "127.0.0.1";
                    options.port = 8888;
                }

                var buffers = [];
                var http = require(url.split(":")[0]);
                var zlib = require("zlib");
                var req = http.request(options, function (res) {
                    res.on("data", function (chunk) {
                        buffers[buffers.length] = chunk;
                    });
                    res.on("end", function () {
                        var response = Buffer.concat(buffers);
                        var callbackWrapper = function (err, response) {
                            callback({ statusCode: res.statusCode, headers: res.headers, body: response.toString("utf8") });
                        }
                        switch (res.headers["content-encoding"]) {
                            case "gzip":
                                zlib.gunzip(response, callbackWrapper);
                                break;
                            case "deflate":
                                zlib.inflate(response, callbackWrapper);
                                break;
                            default:
                                callbackWrapper(undefined, response);
                                break;
                        }
                    });
                    res.on("close", function (err) { engine.Log(err); });
                });
                req.on("error", function (err) { engine.Log(err); callback({ statusCode: 500, body: err.message }); });
                req.write(requestBody, "utf8");
                req.end();
            } else {
                // Browser
                /*global XMLHttpRequest: true*/
                var http = new XMLHttpRequest();
                http.open(method, url, true);
                if (headers) {
                    headers._forEach(function (k, v) { http.setRequestHeader(k, v); });
                }
                http.onreadystatechange = function () {
                    if (http.readyState === http.DONE) {
                        var headerMap = {};
                        http.getAllResponseHeaders().split("\r\n")._forEach(function (k, v) {
                            if (v.length > 0) {
                                var index = v.indexOf(": ");
                                headerMap[v.substring(0, index).toLowerCase()] = v.substring(index + 1);
                            }
                        });
                        callback({ statusCode: http.status, headers: headerMap, body: http.responseText });
                    }
                };
                http.send(requestBody);
            }
        };
        
        this.Log = log.log;
        this.LogFormat = log.toString;
        this.Assert = function (test, message) {
            if (!test) {
                this.Log(message);
                throw message;
            }
        };
        
        var currentTime = new Date().getTime();
        function Now() {
            var dateNow = new Date();
            if (dateNow.valueOf() < currentTime + 2) {
                dateNow.setTime(currentTime + 2);
            }
            currentTime = dateNow.getTime();
            return dateNow;
        };
        
        this._extends("Callbacks", { value: [] });
        
        // _mode: used for indicating whether to call callbacks
        //        1 - call for all changes
        //        2 - call only for user changes (_statementsWriter)
        //        undefined - expected to be only for partner callbacks, will be called only for user changes.        
        this._extends("Notify", this._thunk(function (s, p, o, isCreate) {
            this.Callbacks.map(this._thunk(function (callback) {
                if ((!callback._mode && !this._recordDelta) ||
                    (callback._mode === 2 && this._recordDelta) ||
                    (callback._mode === 1)) {
                    callback(s, p, o, isCreate);
                }
            }))
        }));
        
        this._extends("_runWithoutDeltaRecording", this._thunk(function (func) {
            var prevRecordDelta = this._recordDelta;
            try {
                this._recordDelta = false;
                func();
            } finally {
                this._recordDelta = prevRecordDelta;
            }
        }));
        
        var namespaces = new tokenizer.MapNamespaces(this, true);
        this._extends("_statements", { value: [] });
        
        var statementCallback = this._thunk(function (s, p, o, create) {
            if (!s.IsSubPathOf(this.Entities.Me.Id) ||
                s.IsSubPathOf(this.Entities.MyStatement.Id)) {
                return;
            }
            
            if (this._statements.length === 0) {
                this._statements.push(namespaces.TurtleHeader);
            }
            
            this._statements.push("p:me/statement." + (100000 + this._statements.length) + " a rdf:Statement" + 
                                  " ; rdf:subject " + namespaces.Compress(s.Id) + 
                                  " ; rdf:predicate " + namespaces.Compress(p.Id) +
                                  " ; rdf:object " + namespaces.Compress(o.Id) + 
                                  (create ? " ; s:dateCreated \"" : " ; s:dateDeleted \"") + Now().toISOString() + "\"^^s:Date .\r\n");
        });
        statementCallback._mode = 2;
        
        this.Callbacks.push(statementCallback);

        if (callback) {
            this.Callbacks.push(callback);
        }
        
        this.Create = function (id) {
            switch (typeof (id)) {
                case "string": break;
                case "number": id = id.toString(); break;
                case "boolean": id = id.toString(); break;
                case "object":
                    if (typeof (id.toISOString) === "function") {
                        id = '"' + id.toISOString() + '"^^' + this.Entities.Date.Id;
                    } else {
                        throw "Invalid object";
                    }
                    break;
            }

            var term = this._mapCache[id];
            
            if (!term) {
                term = roots._any(function (k, v) { return v.Create(id); });
                this._mapCache[id] = term;
            }
            
            return term;
        };
        
        this.CreateGeometryPoint = function (latitude, longitude, altitude, scale) {
            var pointText = latitude + " " + longitude;
            var altitudeIsDefined = typeof (altitude) !== "undefined";
            var scaleIsDefined = typeof (scale) !== "undefined";
            var altScaleText = altitudeIsDefined && scaleIsDefined ? " " + altitude + " " + scale :
                                                 altitudeIsDefined ? " " + altitude :
                                                    scaleIsDefined ? " NULL " + scale : "";
            return this.Create('"Point(' + pointText + altScaleText + ')"^^' + this.Entities.Geometry);
        };

        this.CreateText = function (text, culture) {
            // TODO: Find a better way to locate Text entry in roots.
            return this.Create(roots[6].Normalize(text, culture));
        };
        
        this.CreateTyped = function (type, relId) {
            var typeId = type.Id;
            var ending = typeId._findLast("#");
            if (ending.length == 0) {
                ending = typeId._findLast("/");
            }
            
            var cls = ending.substring(1, 2).toLowerCase() + ending.substring(2, ending.length - 1);
            if (!relId) {
                relId = String._createGuid();
            }

            var id = "<http://platform.bing.com/persons/me/" + cls + "." + relId + ">";

            var entity = this.Create(id);
            entity.Set(this.Entities.Type, type);
            return entity;
        };

        // Setup prototype hierarchy
        var termsRoot = Terms.call({}, this);
        
        var termRoot = Term.call({}, this, termsRoot);
        var roots = [
            Entity,
            Boolean,
            Integer, 
            Double, 
            RDFDate,
            GeometryPoint, 
            Text, 
            Variable,
            Symbol].map(function (c) {
            return c.call(termRoot._createChild());
        });
        
        this._extends("Entities", { value: {} });
        enums._forEach(this._thunk(function (k, v) {
            this.Entities[k] = this.Create(v);
        }));
        
        var entityRoot = roots[0];
        var geoAnnotationRoot = GeoAnnotation.call(entityRoot._createChild());
        var t = { "Person": Person,
                  "PreferencesType": Preferences,
                  "FavoritesType": Favorites,
                  "PostalAddress": PostalAddress,
                  "GeoAnnotationCollection": GeoAnnotationCollection
        }._forEach(this._thunk(function (k, v) {
            this.Entities[k]._extends("_wrap", {
                value: v.call(entityRoot._createChild())
            });
        }));

        t = { "Place": Place,
              "Route": Route
        }._forEach(this._thunk(function (k, v) {
            this.Entities[k]._extends("_wrap", {
                value: v.call(geoAnnotationRoot._createChild())
            });
        }));
        
        [ "BluetoothIdentity", 
          "FavoritesSearch",
          "LocalScoutCategory",
          "Location",
          "MemberOf",
          "Members",
          "WayPoint",
          "WifiIdentity",
          "WRPMigratedFavorites",
          "InternalDiagnosticsMessage"]._forEach(this._thunk(function (k, v) {
            this.Entities[v]._extends("IsMultiValued", { value: true });
        }));
        
        // 
        
        this._extends("ImportTurtle", function (data) {
            this.Assert(typeof(data) === "string" && data.length > 0, "ImportTurtle requires a non-empty string.")

            this._runWithoutDeltaRecording(this._thunk(function () {
                var namespaces = new tokenizer.MapNamespaces(this);
                var tokens = new tokenizer.Tokenizer(this, data);
                
                var subject;
                var predicate;
                var object;
                var reuse = 0;
                while (true) {
                    var token;
                    if (reuse < 1) {
                        token = tokens.GetNextToken();
                        if (!token.value) {
                            break;
                        }
                        subject = this.Create(namespaces.Expand(token.value));
                    }
                    
                    if (reuse < 2) {
                        token = tokens.GetNextToken();
                        predicate = this.Create(namespaces.Expand(token.value));
                    }
                    
                    token = tokens.GetNextToken();
                    object = this.Create(namespaces.Expand(token.value));
                    
                    token = tokens.GetNextToken();
                    switch (token.value) {
                        case ".": reuse = 0; break;
                        case ";": reuse = 1; break;
                        case ",": reuse = 2; break;
                        default: throw "ParseError " + token.value;
                    }
                    
                    if (subject.Id == "@prefix") {
                        namespaces.ExtendNamespace(predicate.Id, object.Id.substring(1, object.Id.length - 1));
                    } else {
                        var term = this.Create(subject.Id);
                        term.Out(predicate).Append(object);
                        
                        if (reuse == 0 && term && term.Type == this.Entities.Statement) {
                            var sub = term.Get(this.Entities.Subject);
                            var pre = term.Get(this.Entities.Predicate);
                            var obj = term.Get(this.Entities.Object);
                            var cre = term.Get(this.Entities.DateCreated);
                            if (cre) {
                                sub.Out(pre).Append(obj);
                            } else {
                                sub.Out(pre).Remove(obj);
                            }
                        }
                    }
                }
            }));
        });
        
        this._extends("Sync", function (callback, rpsToken, cookies) {
            var headers = {
                Accept: "*/*", 
                ETag: this._etag,
                "X-CG-Version": "ClientGraph/1.0 (JavaScript 1.0)",
                "Accept-Encoding": "gzip, deflate",
                "Cache-Control": "no-cache",
                "Content-Type": "text/turtle",
                "X-Search-RPSToken" : rpsToken
            };
            
            if (cookies) {
                headers["Cookie"] = cookies;
            }
            
            // Only upload a certain number of statements to avoid going past the 200k upload limit
            var numStatements = Math.min(this._statements.length, 500);
            var payload = this._statements.join("");
            if (payload.length === 0) {
                payload = " ";
            }

            // TODO: Convert to https before release.
            var host = debug ? "http://ssl.bing.com" : "https://ssl.bing.com";
            http.call(this, host + "/CloudGraph/v1/Sync", "POST", headers, payload, this._thunk(function (response) {
                if (response.statusCode === 400) {
                    // Bad input, data is wiped so that it's not re-transmitted.  Notify the caller.
                    this._statements.splice(0, numStatements);
                } else if (response.statusCode !== 200) {
                    // Leave the original statements for another attempt at upload.
                } else {
                    this._statements.splice(0, numStatements);
                    this.ImportTurtle(response.body);
                    this._etag = response.headers["etag"];
                }

                if (callback) {
                    callback(response.statusCode);
                }
            }));
        });

        this.ImportTurtle(config);

        this._extends("Me", { value: this.Entities.Me.Wrap });

        this.Log({ "#green": "Client Graph initialized." });
    }
    
    //     _____
    //    |_   _|__ _ __ _ __ ___
    //      | |/ _ \ '__| '_ ` _ \
    //      | |  __/ |  | | | | | |
    //      |_|\___|_|  |_| |_| |_|
    //    
    function Term(graph, termsRoot) {
        this._extends("_init", function (id) {
            graph.Assert(typeof (id) === "string" || id.IsTerm, "TermProto: Invalid arg");
            
            if (id.IsTerm) {
                this._extends("Id", { value: id.Id, enumerable: true });
                this._extends("_InMap", { value: id._InMap });
                this._extends("_OutMap", { value: id._OutMap });
            } else {
                this._extends("Id", { value: id, enumerable: true });
                this._extends("_InMap", { value: id._startsWith("<") ? {} : undefined });
                this._extends("_OutMap", { value: {} });
            }
            return this;
        });
        
        this._extends("Graph", { value: graph });
        this._extends("IsTerm", { value: true });
        this._extends("IsValid", { get: function () { return this.Id.length > 0; } });

        this._extends("_isOrphaned", {
            get: function () {
                if (!this._InMap) {
                    return true;
                }
                
                return this._InMap._all(this._thunk(function (k, v) {
                    return k === this.Graph.Entities.Statement || v.Values._keys.length === 0;
                }));
            }
        });
        
        this._extends("IsSubPathOf", function (id) {
            graph.Assert(typeof (id) === "string", "Term: Invalid arg");
            return this.Id._startsWith(id.slice(0, -1));
        });
        
        this._extends("toString", function () { return this.Id; });
        
        this._extends("InKeys", {
            get: function () {
                return this._InMap ? this._InMap._keys.map(function (v) { return graph.Create(v); }) : [];
            }
        });
        
        this._extends("OutKeys", {
            get: function () {
                return this._OutMap._keys.map(function (v) { return graph.Create(v); });
            }
        });
        
        this.In = function (predicate) {
            if (!this._InMap) {
                return;
            }
            var terms = this._InMap[predicate.Id];
            if (!terms) {
                this._InMap[predicate.Id] = terms = termsRoot._createChild(this, predicate, false);
            }
            return terms;
        };
        
        this.Out = function (predicate) {
            predicate = predicate.IsTerm ? predicate : graph.Create(predicate);
            var terms = this._OutMap[predicate.Id];
            if (terms === undefined) {
                this._OutMap[predicate.Id] = terms = termsRoot._createChild(this, predicate, true);
            }
            return terms;
        };
        
        this.Get = function (id) {
            return this.Out(id).FirstOrDefault;
        }
        
        this.Set = function (id, term) {
            this.Out(id).Append(term);
        }
        
        this._extends("_createAccessors", function (name, predicate) {
            predicate = predicate || this.Graph.Entities[name];
            this._extends(name, {
                get: function () {
                    var ret = this.Get(predicate);
                    return ret ? ret.Wrap : ret;
                },
                set: function (value) {
                    this.Set(predicate, value ? value.Unwrap : value);
                }
            });
        });
        
        this._extends("_createAccessorForType", function (name) {
            this._extends("Type", {
                get: function () {
                    // Delay memoizing since graph is not bootstrapped yet.
                    var value = this.Graph.Entities[name];
                    this._parent._extends("Type", { value: value });
                    return value;
                },
                configurable: true
            });
        });
        
        this._extends("Wrap", { get: function () { return this; }, configurable: true });
        this._extends("Unwrap", { get: function () { return this; }, configurable: true });

        this.Create = function (id, derived) {
            return this._init.call(derived || this._createChild(), id);
        }
        
        this.Delete = function () {
            this.InKeys._forEach(this._thunk(function (k, v) {
                this._InMap[v].Clear();
            }));
            this.OutKeys._forEach(this._thunk(function (k, v) {
                this._OutMap[v].Clear();
            }));
        }
        
        return this;
    }
    
    //      ____              _
    //     | __ )  ___   ___ | | ___  __ _ _ __
    //     |  _ \ / _ \ / _ \| |/ _ \/ _` | '_ \
    //     | |_) | (_) | (_) | |  __/ (_| | | | |
    //     |____/ \___/ \___/|_|\___|\__,_|_| |_|
    //
    function Boolean() {
        function init(id, value) {
            this._init(id);
            this._extends("Value", { value: value });
            return this;
        }
        
        this._createAccessorForType("Boolean");
        
        this.Create = function (id) {
            if (id === "true" || id === "false") {
                return init.call(this._createChild(), id, id === "true");
            }
        }
        
        return this;
    }

    //     ___       _
    //    |_ _|_ __ | |_ ___  __ _  ___ _ __
    //     | || '_ \| __/ _ \/ _` |/ _ \ '__|
    //     | || | | | ||  __/ (_| |  __/ |
    //    |___|_| |_|\__\___|\__, |\___|_|
    //                       |___/
    function Integer() {
        function init(id, value) {
            this._init(id);
            this._extends("Value", { value: value });
            return this;
        }
        
        this._createAccessorForType("Integer");
        
        this.Create = function (id) {
            var value = Number(id);
            if (!isNaN(value) && id.indexOf(".") === -1) {
                return init.call(this._createChild(), id, value);
            }
        }
        
        return this;
    }
    
    //      ____              _     _
    //     |  _ \  ___  _   _| |__ | | ___
    //     | | | |/ _ \| | | | '_ \| |/ _ \
    //     | |_| | (_) | |_| | |_) | |  __/
    //     |____/ \___/ \__,_|_.__/|_|\___|
    //
    function Double() {
        function init(id, value) {
            this._init(id);
            this._extends("Value", { value: value });
            return this;
        }
        
        this._createAccessorForType("Double");
        
        this.Create = function (id) {
            var value = Number(id);
            if (!isNaN(value)) {
                return init.call(this._createChild(), id, value);
            }
        }
        
        return this;
    }
    
    //      ____        _
    //     |  _ \  __ _| |_ ___
    //     | | | |/ _` | __/ _ \
    //     | |_| | (_| | ||  __/
    //     |____/ \__,_|\__\___|
    //
    function RDFDate() {
        function init(id, value) {
            this._init(id);
            this._extends("Value", { value: value });
            return this;
        }
        
        this._createAccessorForType("Date");
        
        this.Create = function (id) {
            if (id._startsWith('"') && id._endsWith("^^" + this.Graph.Entities.Date.Id)) {
                var inner = id.split('"')[1];
                var value = new Date(inner);
                if (!isNaN(value.valueOf())) {
                    return init.call(this._createChild(), id, value);
                }
            }
        }
        
        return this;
    }
    
    //      _____         _
    //     |_   _|____  _| |_
    //       | |/ _ \ \/ / __|
    //       | |  __/>  <| |_
    //       |_|\___/_/\_\\__|
    //
    function Text() {
        function init(id, value) {
            this._init(id);
            this._extends("Value", { value: value });
            return this;
        }
        
        this._createAccessorForType("Text");
        
        this.Normalize = function (data, culture) {
            var accuml = '"';
            var len = data.length;
            for (var i = 0; i < len; ) {
                var ch = data[i++];
                switch (ch) {
                    case '"':
                    case "'":
                    case "\\":
                    case "/":
                        accuml += "\\" + ch;
                        break;
                    case "\b":
                        accuml += "\\b";
                        break;
                    case "\f":
                        accuml += "\\f";
                        break;
                    case "\n":
                        accuml += "\\n";
                        break;
                    case "\r":
                        accuml += "\\r";
                        break;
                    case "\t":
                        accuml += "\\t";
                        break;
                    default:
                        if (0x20 <= ch.charCodeAt(0) && ch.charCodeAt(0) <= 0x7f) {
                            accuml += ch;
                        } else {
                            var hex = ch.charCodeAt(0).toString(16);
                            accuml += "\\U" + "0000".substring(hex.length) + hex;
                        }
                        break;
                }
            }
            accuml += '"';
            if (culture) {
                accuml += "@" + culture;
            }
            return accuml;
        }
        
        this.Create = function (id) {
            var quote = id[0];
            if (quote !== '\"') {
                return;
            }
            
            // TODO: Use stringbuilder
            var accuml = "";
            var i = 1;
            var len = id.length;
            
            var tripleQuote = len >= 6 && id[1] == quote && id[2] == quote;
            if (tripleQuote) {
                i += 2;
            }
            
            while (true) {
                var ch = id[i++];
                this.Graph.Assert((0x20 <= ch.charCodeAt(0) && ch.charCodeAt(0) <= 0x7f) || (tripleQuote && (ch == '\r' || ch == '\n')), "Text: Invalid char " + ch);
                
                if (ch == '\\') {
                    ch = id[i++];
                    switch (ch) {
                        case '"':
                        case '\'':
                        case '\\':
                        case '/': accuml += ch; break;
                        case 'b': accuml += '\b'; break;
                        case 'f': accuml += '\f'; break;
                        case 'n': accuml += '\n'; break;
                        case 'r': accuml += '\r'; break;
                        case 't': accuml += '\t'; break;
                        case 'u':
                        case 'U':
                            var u = Number("0x" + id.substring(i, i + 4));
                            this.Graph.Assert(!isNaN(u));
                            accuml += String.fromCharCode(u);
                            i += 4;
                            break;
                    }
                } else if (ch == quote) {
                    if (tripleQuote && id[i] == quote && id[i + 1] == quote) {
                        i += 2;
                        break;
                    } else if (!tripleQuote) {
                        break;
                    } else {
                        accuml += ch;
                    }
                } else {
                    accuml += ch;
                }
            }
            
            // TODO: Culture.
            return init.call(this._createChild(), id, accuml);
        }
        return this;
    }
    
    //     __     __         _       _     _
    //     \ \   / /_ _ _ __(_) __ _| |__ | | ___
    //      \ \ / / _` | '__| |/ _` | '_ \| |/ _ \
    //       \ V / (_| | |  | | (_| | |_) | |  __/
    //        \_/ \__,_|_|  |_|\__,_|_.__/|_|\___|
    //
    function Variable() {
        function init(id) {
            this._init(id);
            return this;
        }
        
        this._createAccessorForType("Variable");
        
        this.Create = function (id) {
            if (id._startsWith("?") || id._startsWith("$")) {
                return init.call(this._createChild(), id);
            }
        }
        
        return this;
    }
    
    //      ____                  _           _
    //     / ___| _   _ _ __ ___ | |__   ___ | |
    //     \___ \| | | | '_ ` _ \| '_ \ / _ \| |
    //      ___) | |_| | | | | | | |_) | (_) | |
    //     |____/ \__, |_| |_| |_|_.__/ \___/|_|
    //            |___/
    //
    function Symbol() {
        function init(id) {
            this._init(id);
            return this;
        }
        
        this._createAccessorForType("Symbol");
        
        this.Create = function (id) {
            return init.call(this._createChild(), id);
        }
        
        return this;
    }
    
    //       ____                           _              ____       _       _
    //      / ___| ___  ___  _ __ ___   ___| |_ _ __ _   _|  _ \ ___ (_)_ __ | |_
    //     | |  _ / _ \/ _ \| '_ ` _ \ / _ \ __| '__| | | | |_) / _ \| | '_ \| __|
    //     | |_| |  __/ (_) | | | | | |  __/ |_| |  | |_| |  __/ (_) | | | | | |_
    //      \____|\___|\___/|_| |_| |_|\___|\__|_|   \__, |_|   \___/|_|_| |_|\__|
    //                                               |___/
    //
    function GeometryPoint() {
        function init(id, coordinates) {
            this._init(id);
            this._extends("Latitude", { value: coordinates[0] });
            this._extends("Longitude", { value: coordinates[1] });
            this._extends("Altitude", { value: coordinates[2] });
            this._extends("Scale", { value: coordinates[3] });
            return this;
        }
        
        this._createAccessorForType("Geometry");
        
        this.Create = function (id) {
            if (id._startsWith('"Point(') && id._endsWith(')"^^' + this.Graph.Entities.Geometry.Id)) {
                var inner = id._replace('"Point(', '')._replace(')"^^' + this.Graph.Entities.Geometry.Id, '');
                
                var coordinates = [];
                inner.split(" ")._forEach(this._thunk(function (k, v) {
                    if (v.length === 0) {
                        // Skip
                    } else if (v === "NULL") {
                        coordinates.push(null);
                    } else {
                        var coord = this.Graph.Create(v);
                        this.Graph.Assert(coord.Type === this.Graph.Entities.Integer || coord.Type === this.Graph.Entities.Double, "GeometryPoint expected a number");
                        coordinates.push(coord);
                    }
                }));
                
                this.Graph.Assert(coordinates.length >= 2 && coordinates.length <= 4 && coordinates[0] && coordinates[1], "Invalid GeometryPoint coordinates");
                return init.call(this._createChild(), id, coordinates);
            }
        }
        
        return this;
    }

    //       ____                           _              _     _
    //      / ___| ___  ___  _ __ ___   ___| |_ _ __ _   _| |   (_)_ __   ___
    //     | |  _ / _ \/ _ \| '_ ` _ \ / _ \ __| '__| | | | |   | | '_ \ / _ \
    //     | |_| |  __/ (_) | | | | | |  __/ |_| |  | |_| | |___| | | | |  __/
    //      \____|\___|\___/|_| |_| |_|\___|\__|_|   \__, |_____|_|_| |_|\___|
    //                                               |___/
    //

    //       ____                           _              ____       _
    //      / ___| ___  ___  _ __ ___   ___| |_ _ __ _   _|  _ \ ___ | |_   _  __ _  ___  _ __
    //     | |  _ / _ \/ _ \| '_ ` _ \ / _ \ __| '__| | | | |_) / _ \| | | | |/ _` |/ _ \| '_ \
    //     | |_| |  __/ (_) | | | | | |  __/ |_| |  | |_| |  __/ (_) | | |_| | (_| | (_) | | | |
    //      \____|\___|\___/|_| |_| |_|\___|\__|_|   \__, |_|   \___/|_|\__, |\__, |\___/|_| |_|
    //                                               |___/              |___/ |___/
    //

    //      _____
    //     |_   _|__ _ __ _ __ ___  ___
    //       | |/ _ \ '__| '_ ` _ \/ __|
    //       | |  __/ |  | | | | | \__ \
    //       |_|\___|_|  |_| |_| |_|___/
    //
    function Terms(graph) {
        this._extends("_initChild", function (parent, predicate, isOut) {
            this._extends("Values", { value: {} });
            this._extends("Parent", { value: parent });
            this._extends("Predicate", { value: predicate });
            this._extends("IsOut", { value: isOut });
        });
        
        this._extends("Graph", { value: graph });
        
        this._extends("Append", function (term) {
            var term = term.Unwrap;

            if (!this.IsOut) {
                return term.Out(this.Predicate).Append(this.Parent);
            }
            
            if (term && this.Exists(term)) {
                return;
            }
            
            if (!this.Predicate.IsMultiValued) {
                this.Values._clear();
            }

            if (term) {
                this._Add(term);

                var inTerms = term.In(this.Predicate);
                if (inTerms) {
                    inTerms._Add(this.Parent);
                }
            }
            
            graph.Notify(this.Parent, this.Predicate, term, true);
        });
        
        this._extends("Remove", function (term) {
            var term = term.Unwrap;

            if (!this.IsOut) {
                return term.Out(this.Predicate).Remove(this.Parent);
            }

            if (!term || !this.Exists(term)) {
                return;
            }

            this._Remove(term);
            var inTerms = term.In(this.Predicate);
            if (inTerms) {
                inTerms._Remove(this.Parent);
            }

            graph.Notify(this.Parent, this.Predicate, term, false);

            if (term._isOrphaned) {
                term.Delete();
            }
        });
        
        this._extends("Clear", function () {
            this.Values._forEach(this._thunk(function (k, v) {
                this.Remove(v);
            }));
        });

        this._extends("Exists", function (term) {
            return this.Values[term.Id] !== undefined;
        });

        this._extends("FirstOrDefault", {
            get: function () {
                return this.Values._any(function (k, v) {
                    return v;
                });
            }
        });

        this._extends("_Add", function (term) {
            this.Values[term.Id] = term;
        });

        this._extends("_Remove", function (term) {
            delete this.Values[term.Id];
        });
        
        return this;
    }

    //      _     _     _
    //     | |   (_)___| |_
    //     | |   | / __| __|
    //     | |___| \__ \ |_
    //     |_____|_|___/\__|
    //
    //               Parent_Entity
    //                     ^
    //                     | memberOf
    //      List-Node-0         List-Node-1         ...
    //         | rdf:first           | rdf:first
    //         V                     V
    //       term                  term
    function List(terms) {
        var list = [];
        var actual = [];
        var isDirty = true;
        
        this.Graph = terms.Graph;
        
        var callback = function (s, p, o, create) {
            isDirty = isDirty || p === terms.Predicate;
        };

        callback._mode = 1;
        this.Graph.Callbacks.push(callback);
        
        function RebuildCache() {
            if (!isDirty) {
                return;
            }
            isDirty = false;
            
            list = [];
            actual = [];
            
            // TODO: We may need to re-add displayOrder setting based on #Origin/#Destination suffixes if the client passes them in.
            var mapTerms = {};
            var maxTerms = 0;
            terms.Values._forEach(this._thunk(function (k, v) {
                var first = v.Get(this.Graph.Entities.First);
                if (first) {
                    var displayOrder = v.Get(this.Graph.Entities.DisplayOrder);
                    var num = displayOrder ? displayOrder.Value : maxTerms;
                    if (mapTerms[num]) {
                        num = maxTerms + 1;
                    }
                    mapTerms[num] = v;
                    maxTerms = num > maxTerms ? num : maxTerms;
                }
            }));
            
            mapTerms._forEach(this._thunk(function (k, v) {
                list.push(v);
                actual.push(v.Get(this.Graph.Entities.First).Wrap);
            }));
        }
        
        this._extends("Count", {
            get: function () {
                RebuildCache.call(this);
                return actual.length;
            }
        });
        
        this._extends("IndexOf", function (term) {
            RebuildCache.call(this);
            return actual._any(function (k, v) {
                if (v === term) {
                    return k;
                }
            });
        });
        
        this._extends("GetAt", function (index) {
            RebuildCache.call(this);
            return actual[index];
        });
        
        this._extends("SetAt", function (index, item) {
            RebuildCache.call(this);
            
            var listNode = list[index];
            var rawItem = item.Unwrap;
            listNode.Set(this.Graph.Entities.DisplayOrder, this.Graph.Create(index));
            listNode.Set(this.Graph.Entities.First, rawItem);
            actual[index] = item;
        });
        
        this._extends("InsertAt", function (index, item) {
            RebuildCache.call(this);
            
            var term = item.Unwrap;
            var id = term.Id._findLast("/")._find(".");
            if (id.length) {
                id = id.substring(1, id.length - 1);
            }
            if (terms.Parent.Unwrap !== this.Graph.Entities.FavoritesDefaultCollection) {
                // If non default collection we decorate name with collection 
                // name to avoid clashes if item in multiple collections.
                var idPrefix = terms.Parent.Id;
                idPrefix = idPrefix.substring(0, idPrefix.length - 1);
                var lastTrail = idPrefix._findLast("/");
                idPrefix = lastTrail._find(".");
                idPrefix = idPrefix.length ? idPrefix.substring(1) : lastTrail;
                id = idPrefix + "." + id;
            }
            if (this.Graph._mapCache["<http://platform.bing.com/persons/me/list." + id + ">"]) {
                id = undefined;
            }
            
            // Fixup displayOrder
            for (var i = index; i < list.length; i++) {
                var listNode = list[i];
                if (listNode) {
                    listNode.Set(this.Graph.Entities.DisplayOrder, this.Graph.Create(i + 1));
                }
            }
            
            // Create a new node
            var listNode = this.Graph.CreateTyped(this.Graph.Entities.List, id);
            listNode.Set(this.Graph.Entities.First, term);
            listNode.Set(this.Graph.Entities.DisplayOrder, this.Graph.Create(index));
            
            // This will trigger term property change
            terms.Append(listNode);
        });
        
        this._extends("Append", function (item) {
            RebuildCache.call(this);
            this.InsertAt(list.length, item);
        });
        
        this._extends("Remove", function (item) {
            RebuildCache.call(this);
            var index = this.IndexOf(item);
            if (index >= 0) {
                this.RemoveAt(index);
            }
        });
        
        this._extends("RemoveAt", function (index) {
            RebuildCache.call(this);
            
            var listNode = list[index];
            if (!listNode) {
                return;
            }
            
            var first = listNode.Get(this.Graph.Entities.First);
            if (first) {
                // If removing from the "Default" favorites, ensure item erased 
                // from all Preferred "in"
                if (terms.Parent.Unwrap === this.Graph.Entities.FavoritesDefaultCollection) {
                    var cat = this.Graph.Entities.FavoritesCategory;
                    first.InKeys._forEach(function (k, v) {
                        if (v.IsSubPathOf(cat.Id)) {
                            first.In(v).Clear();
                        }
                    });
                }
            }
            listNode.Delete();
        });

        this._extends("CreateItem", function (typeId, relId) {
            var newItem = this.Graph.CreateTyped(typeId, relId).Wrap;
            this.Append(newItem);
            return newItem;
        });
    }
    
    //      _____       _   _ _
    //     | ____|_ __ | |_(_) |_ _   _
    //     |  _| | '_ \| __| | __| | | |
    //     | |___| | | | |_| | |_| |_| |
    //     |_____|_| |_|\__|_|\__|\__, |
    //                            |___/
    function Entity() {
        
        function init(id) {
            this._init(id);
            return this;
        }
        
        this._extends("_initChild", function (term) {
            if (term && term.IsTerm) {
                this._init(term);
            }
        });
        
        this._extends("IsEntity", { value: true });
        
        this._extends("Type", {
            get: function () {
                return this.Get(this.Graph.Entities.Type);
            },
            configurable: true
        });
        
        this.Create = function (id) {
            if (id._startsWith("<")) {
                return init.call(this._createChild(), id);
            }
        };
        
        this._extends("Wrap", {
            get: function () {
                if (this._wrapped) {
                    return this._wrapped;
                }
                var derived = this.Type;
                if (derived && derived._wrap) {
                    var ret = this._wrapped = derived._wrap._createChild(this);
                    ret._unwrapped = this;
                    return ret;
                }
                return this;
            }
        });
        
        this._extends("Unwrap", {
            get: function () {
                return this._unwrapped || this;
            }
        });
        
        return this;
    }
    
    //      ____
    //     |  _ \ ___ _ __ ___  ___  _ __
    //     | |_) / _ \ '__/ __|/ _ \| '_ \
    //     |  __/  __/ |  \__ \ (_) | | | |
    //     |_|   \___|_|  |___/\___/|_| |_|
    //
    function Person() {
        this._createAccessors("Preferences");
        return this;
    }
    
    //      ____            __
    //     |  _ \ _ __ ___ / _| ___ _ __ ___ _ __   ___ ___  ___
    //     | |_) | '__/ _ \ |_ / _ \ '__/ _ \ '_ \ / __/ _ \/ __|
    //     |  __/| | |  __/  _|  __/ | |  __/ | | | (_|  __/\__ \
    //     |_|   |_|  \___|_|  \___|_|  \___|_| |_|\___\___||___/
    //
    function Preferences() {
        this._createAccessors("Favorites");
        return this;
    }
    
    //      _____                     _ _
    //     |  ___|_ ___   _____  _ __(_) |_ ___  ___
    //     | |_ / _` \ \ / / _ \| '__| | __/ _ \/ __|
    //     |  _| (_| |\ V / (_) | |  | | ||  __/\__ \
    //     |_|  \__,_| \_/ \___/|_|  |_|\__\___||___/
    //
    function Favorites() {
        this._extends("GetPreferredPlace", function (term) {
            var place = this.Get(term);
            if (place) {
                return this.Get(term).Wrap;
            }
        });
        
        this._extends("SetPreferredPlace", function (favoritesCategory, place) {
            this.Set(favoritesCategory, place);
        });

        this._extends("Collections", {
            get: function () {
                // Delay memoizing since graph is not bootstrapped yet.
                var value = new List(this.In(this.Graph.Entities.MemberOf));
                this._extends("Collections", { value: value });
                
                // Extract Default and Recent
                value._extends("Default", { value: this.Graph.Entities.FavoritesDefaultCollection.Wrap });
                value._extends("Recent", { value: this.Graph.Entities.FavoritesRecentCollection.Wrap });
                
                return value;
            },
            configurable: true
        });
        
        return this;
    }
   
    //        ____              _                      _        _   _              ____      _ _           _   _
    //       / ___| ___  ___   / \   _ __  _ __   ___ | |_ __ _| |_(_) ___  _ __  / ___|___ | | | ___  ___| |_(_) ___  _ __
    //      | |  _ / _ \/ _ \ / _ \ | '_ \| '_ \ / _ \| __/ _` | __| |/ _ \| '_ \| |   / _ \| | |/ _ \/ __| __| |/ _ \| '_ \
    //      | |_| |  __/ (_) / ___ \| | | | | | | (_) | || (_| | |_| | (_) | | | | |__| (_) | | |  __/ (__| |_| | (_) | | | |
    //       \____|\___|\___/_/   \_\_| |_|_| |_|\___/ \__\__,_|\__|_|\___/|_| |_|\____\___/|_|_|\___|\___|\__|_|\___/|_| |_|
    //
    function GeoAnnotationCollection() {
        this._extends("Members", {
            get: function () {
                var value = new List(this.In(this.Graph.Entities.MemberOf));
                this._extends("Members", { value: value });
                return value;
            },
            configurable: true
        });
          
        this._extends("CreatePlace", function (relId) {
            return this.Members.CreateItem(this.Graph.Entities.Place, relId);
        });
        
        this._extends("CreateRoute", function (relId) {
            return this.Members.CreateItem(this.Graph.Entities.Route, relId);
        });

        return this;
    }
    
    //        ____              _                      _        _   _
    //       / ___| ___  ___   / \   _ __  _ __   ___ | |_ __ _| |_(_) ___  _ __
    //      | |  _ / _ \/ _ \ / _ \ | '_ \| '_ \ / _ \| __/ _` | __| |/ _ \| '_ \
    //      | |_| |  __/ (_) / ___ \| | | | | | | (_) | || (_| | |_| | (_) | | | |
    //       \____|\___|\___/_/   \_\_| |_|_| |_|\___/ \__\__,_|\__|_|\___/|_| |_|
    //
    function GeoAnnotation() {
        // Base type, derived object (Place/Route) provides Type.
        this._createAccessors("BingEntityId");
        this._createAccessors("DateCreated");
        this._createAccessors("DateModified");
        this._createAccessors("DateAccessed");
        this._createAccessors("DisplayOrder");
        this._createAccessors("DisplayCoordinates");
        this._createAccessors("ShowOnMap");
        this._createAccessors("MapZoomLevel");
        this._createAccessors("Name");
        this._createAccessors("Description");
        this._createAccessors("Url");
        this._createAccessors("ImageUri");
        this._createAccessors("Geometry");
        return this;
    }
    
    //      ____  _
    //     |  _ \| | __ _  ___ ___
    //     | |_) | |/ _` |/ __/ _ \
    //     |  __/| | (_| | (_|  __/
    //     |_|   |_|\__,_|\___\___|
    //
    function Place() {
        this._createAccessors("Address");
        this._createAccessors("Phone");
        this._createAccessors("OriginalName");
        this._createAccessors("BusinessType");

        this._extends("IsValid", {
            get: function () {
                return this._parent._parent._parent._parent._property("IsValid").get.call(this) &&
                   this.BingEntityId && this.BingEntityId.Value.length !== 0 &&
                   this.OriginalName && this.OriginalName.Value.length !== 0 &&
                   (this.Geometry || this.DisplayCoordinates) && true; 
            }
        });
        return this;
    }
    
    //      ____           _        _    _       _     _
    //     |  _ \ ___  ___| |_ __ _| |  / \   __| | __| |_ __ ___  ___ ___
    //     | |_) / _ \/ __| __/ _` | | / _ \ / _` |/ _` | '__/ _ \/ __/ __|
    //     |  __/ (_) \__ \ || (_| | |/ ___ \ (_| | (_| | | |  __/\__ \__ \
    //     |_|   \___/|___/\__\__,_|_/_/   \_\__,_|\__,_|_|  \___||___/___/
    //
    function PostalAddress() {
        this._createAccessors("AddressSingleLine");
        this._createAccessors("StreetAddress");
        this._createAccessors("PostalCode");
        this._createAccessors("AddressRegion");
        this._createAccessors("AddressLocality");
        this._createAccessors("AddressCountry");
        return this;
    }
    
    //      ____             _
    //     |  _ \ ___  _   _| |_ ___
    //     | |_) / _ \| | | | __/ _ \
    //     |  _ < (_) | |_| | ||  __/
    //     |_| \_\___/ \__,_|\__\___|
    //
    function Route() {
        this._createAccessors("RouteMode");

        this._extends("Waypoints", {
            get: function () {
                var value = new List(this.Out(this.Graph.Entities.WayPoint));
                this._extends("Waypoints", { value: value });
                return value;
            },
            configurable: true
        });
        
        this._extends("IsValid", {
            get: function () {
                var wpCount = this.Waypoints.Count;
                if (wpCount < 2) {
                    return false;
                }
                
                for (var i = 0; i < wpCount; i++) {
                    if (!this.Waypoints.GetAt(i).IsValid) {
                        return false;
                    }
                }

                return this._parent._parent._parent._parent._property("IsValid").get.call(this) &&
                        this.BingEntityId && this.BingEntityId.Value.length !== 0 &&
                        this.Name && this.Name.Value.length !== 0 && true;
            }
        });
        return this;
    }
    
    //      _____                       _
    //     | ____|_  ___ __   ___  _ __| |_ ___
    //     |  _| \ \/ / '_ \ / _ \| '__| __/ __|
    //     | |___ >  <| |_) | (_) | |  | |_\__ \
    //     |_____/_/\_\ .__/ \___/|_|   \__|___/
    //                |_|
    
    module.exports.CreateGraph = function (callback, debug) {
        return new Graph(callback, debug);
    };
    
    // DO NOT PLACE ANYTHING BELOW THE FOLLOWING LINE -- IT
    // BRACKETS THE ENTIRE FILE.
})();
