// ----------------------------------------------------------------------------------------------------------------------
//   Copyright 2014 (c) Microsoft Corporation.
//   Summary: Examples for API usage.
// ----------------------------------------------------------------------------------------------------------------------
(function () {
    "use strict"; // whole file
    
    //
    //      ____                        _ 
    //     / ___|  __ _ _ __ ___  _ __ | | ___ 
    //     \___ \ / _` | '_ ` _ \| '_ \| |/ _ \
    //      ___) | (_| | | | | | | |_) | |  __/
    //     |____/ \__,_|_| |_| |_| .__/|_|\___|
    //                           |_|           
    //
    function Sample(graph) {
        //
        // Setting a periodic Graph Sync call, every 5 minutes.
        //
        setInterval(function () { graph.Sync(function (statusCode) { graph.Log({ "#green": "Syncing again..." }); }); }, 10 * 60 * 1000);

        //
        // To start, Sync the Graph from the server.
        // 
        // Notes: Sync is an async call.  Include a callback for your custom processing once the graph is synced.
        //
        graph.Sync(function (statusCode) {
            // Note: Be sure to check the status code to see if the Sync was successful.
            if (statusCode !== 200) {
                // Perhaps do something here...
            }

            //
            // Creating a Place
            //
            // Notes: This creates a Place in the Default collection.
            //        At minimum to be valid, a Place requires the following to be set to non-empty values:
            //        BingEntityId, OriginalName, and (Geometry || DisplayCoordinates).
            //
            var newPlace = graph.Me.Preferences.Favorites.Collections.Default.CreatePlace("YN873x112609011");
            graph.Log("Created a Default Place: " + newPlace.Id);
            newPlace.BingEntityId = graph.CreateText("YN873x112609011");
            newPlace.Address = graph.CreateText("400 Broad Street, Seattle, WA, 98109");
            newPlace.Phone = graph.CreateText("(800) 937-9582");
            newPlace.OriginalName = graph.CreateText("Space Needle");
            newPlace.DisplayCoordinates = graph.CreateGeometryPoint(47.620491, -122.349319);
            LogTerm(newPlace);
            graph._setTestStatus(newPlace._isOrphaned === false, "    !_isOrphaned");

            //
            // Getting a Preferred Place
            //
            var home = graph.Me.Preferences.Favorites.GetPreferredPlace(graph.Entities.FavoritesCategoryHome);
            LogTerm(home);

            //
            // Setting a Preferred Place
            //
            graph.Me.Preferences.Favorites.SetPreferredPlace(graph.Entities.FavoritesCategoryWork, newPlace);

            //
            // Removing a Place
            //
            // Notes: Make sure that you Append to another Collection or Route if you don't want to lose the Place's information.
            //        Automatic cleanup is done if a Place is removed and is not attached to anything else in the Graph.
            //        For an example of a "move" operation between Collections, we'll Append the Place to another Collection
            //        before removing it from Default.
            graph.Me.Preferences.Favorites.Collections.Recent.Members.Append(newPlace);
            graph.Me.Preferences.Favorites.Collections.Default.Members.Remove(newPlace);

            //
            // Creating a Route
            //
            // Notes: This creates a Route in the Default collection with two waypoints.
            //        At minimum to be valid, a Route requires the following to be set to non-empty values:
            //        BingEntityId, Name, and two valid Places as Waypoints.
            //
            var newRoute = graph.Me.Preferences.Favorites.Collections.Default.CreateRoute("47.653309_-122.3675~47.640068_-122.12986_Driving_-110409522");
            graph.Log("Created a Default Route: " + newRoute.Id);
            newRoute.BingEntityId = graph.CreateText("47.653309_-122.3675~47.640068_-122.12986_Driving_-110409522");
            newRoute.Name = graph.CreateText("Pacific Science Center to Microsoft Corporation");

            //
            // Adding Waypoints to the Route.
            //
            // Notes: When appending to the route, be sure to Append them in the order of Origin, then Destination.
            //        Notice that the Waypoints are Places created in the Default Collection.  This means they're connected in the Graph
            //        by two entities: Default Collection, and newRoute.
            //
            var origin = graph.Me.Preferences.Favorites.Collections.Default.CreatePlace("47.653309_-122.3675~47.640068_-122.12986_Driving_-110409522#Origin");
            origin.BingEntityId = graph.CreateText("47.653309_-122.3675~47.640068_-122.12986_Driving_-110409522#Origin");
            origin.OriginalName = graph.CreateText("Pacific Science Center");
            origin.DisplayCoordinates = graph.CreateGeometryPoint(47.653309, -122.367500);

            var destination = graph.Me.Preferences.Favorites.Collections.Default.CreatePlace("47.653309_-122.3675~47.640068_-122.12986_Driving_-110409522#Destination");
            destination.BingEntityId = graph.CreateText("47.653309_-122.3675~47.640068_-122.12986_Driving_-110409522#Destination");
            destination.OriginalName = graph.CreateText("Microsoft Corporation");
            destination.DisplayCoordinates = graph.CreateGeometryPoint(47.640068, -122.129860);

            newRoute.Waypoints.Append(origin);
            newRoute.Waypoints.Append(destination);

            //
            // Removing a Route
            //
            // Notes: Since the Waypoints are connected in the Graph to Default Collection and newRoute, they'll
            //        still be available in the Default Collection even after newRoute is Removed.
            //
            graph.Me.Preferences.Favorites.Collections.Default.Members.Remove(newRoute);

            //
            // Removing the Waypoints
            //
            // Notes: To fully remove the Waypoints from the Graph, simple sever the last connection in the Graph to
            //        Default Collection.  This could be done before removing newRoute as well, in which case they
            //        would be automatically cleaned up when newRoute is Removed.
            graph.Me.Preferences.Favorites.Collections.Default.Members.Remove(origin);
            graph.Me.Preferences.Favorites.Collections.Default.Members.Remove(destination);

            graph.Log({ "#green": "Sample Completed." });
        });
    };
    
    function LogTerm(term) {
        if (term === undefined) {
            console.writeline("Term is undefined.");
        } else if (term.Type === term.Graph.Entities.Route) {
            LogRoute(term);
        } else if (term.Type === term.Graph.Entities.Place) {
            LogPlace(term);
        } else if (term.Type === term.Graph.Entities.GeoAnnotationCollection) {
            LogCollection(term);
        }
    }

    function LogCollection(collection) {
        var members = collection.Members;
        for (var i = 0; i < members.Count; i++) {
            var item = members.GetAt(i);
            LogTerm(item);
        }
    };

    function LogPlace(place) {
        if (place === undefined) {
            console.writeline("   Place was undefined.");
        } else {
            var graph = place.Graph;
            graph.Log("    BingEntityId: " + place.BingEntityId);
            graph.Log("    Address: " + place.Address);
            graph.Log("    Phone: " + place.Phone);
            graph.Log("    OriginalName: " + place.OriginalName);
            graph.Log("    BusinessType: " + place.BusinessType);
            if (place.Geometry !== undefined) {
                graph.Log("    Geometry: " + place.Geometry);
            }
            if (place.DisplayCoordinates !== undefined) {
                graph.Log("    DisplayCoordinates: " + place.DisplayCoordinates);
            }
            graph._setTestStatus(place.IsValid, "    isValid");
        }
    };

    function LogRoute(route) {
        if (route === undefined) {
            console.writeline("    Route was undefined.");
        } else {
            var graph = route.Graph;
            graph.Log("    BingEntityId: " + route.BingEntityId);
            graph.Log("    Name: " + route.Name);
            graph.Log("    # Waypoints: " + route.Waypoints.Count);

            var waypoints = route.Waypoints;
            for (var i = 0; i < waypoints.Count; i++) {
                var wp = waypoints.GetAt(i);
                graph.Log("        WayPoints " + i + ": " + wp.Id);
            }
            graph._setTestStatus(route.IsValid, "    isValid");
        }
    };

    //      _____                       _
    //     | ____|_  ___ __   ___  _ __| |_ ___
    //     |  _| \ \/ / '_ \ / _ \| '__| __/ __|
    //     | |___ >  <| |_) | (_) | |  | |_\__ \
    //     |_____/_/\_\ .__/ \___/|_|   \__|___/
    //                |_|
    
    module.exports.Sample = Sample;
    module.exports.LogTerm = LogTerm;
    
    // DO NOT PLACE ANYTHING BELOW THE FOLLOWING LINE -- IT
    // BRACKETS THE ENTIRE FILE.
})();
