// ----------------------------------------------------------------------------------------------------------------------
//   Copyright 2014 (c) Microsoft Corporation.
//   Summary: Defines the Graph Tokenizer.
// ----------------------------------------------------------------------------------------------------------------------
(function () {
    "use strict"; // whole file
    
    function MapNamespaces(graph, init) {
        var namespaces = {};
        var TypeShorthand = "a";

        function InitDefaultsCommon() {
            namespaces["s:"] = "http://schema.org/";
            namespaces["b:"] = "http://platform.bing.com/maps/";
            namespaces["p:"] = "http://platform.bing.com/persons/";
            namespaces["rdf:"] = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
            namespaces["geo:"] = "http://www.opengis.net/ont/geosparql#";
            namespaces["geof:"] = "http://www.opengis.net/ont/geosparql/function/";
        }
        
        function CompressHelper(orig) {
            var len = orig.length;
            if (!orig._startsWith("<") || len < 2) {
                return orig;
            }

            var bare = orig.substr(1, len - 2);
            var lenBare = bare.length;
            return namespaces._any(function (k, v) {
                if (bare._startsWith(v) && lenBare > v.length) {
                    return bare._replace(v, k);
                }
            }) || orig;
        }
        
        function ExpandHelper(orig) {
            var len = orig.length;
            return namespaces._any(function (k, v) {
                if (orig._startsWith(k) && len > k.length) {
                    return "<" + orig._replace(k, v) + ">";
                }
            }) || orig;
        }

        this.Expand = function (orig) {
            if (orig === TypeShorthand) {
                return graph.Entities.Type.Id;
            }
            
            // Embedded type in string
            if (orig._startsWith('"')) {
                var find = orig.lastIndexOf('"^^');
                if (find === -1) {
                    return orig;
                }
                
                find += 3; // "^^
                return orig.substr(0, find) + ExpandHelper(orig.substr(find));
            }
            return ExpandHelper(orig);
        };

        this.Compress = function (orig) {
            if (orig === graph.Entities.Type.Id) {
                return TypeShorthand;
            }

            // Embedded type in string
            if (orig._startsWith('"')) {
                var find = orig.lastIndexOf('"^^<');
                if (find === -1) {
                    return orig;
                }

                find += 3; // "^^
                return orig.substr(0, find) + CompressHelper(orig.substr(find));
            }
            return CompressHelper(orig);
        };

        this._extends("TurtleHeader", {
            get: function () {
                return "@prefix " +
                       namespaces._map(function (k, v) { return k + "<" + v + ">"; }).join(" ; ") +
                       " .\r\n";
            }
        });
        
        this._extends("ExtendNamespace", function (k, v) {
            namespaces[k] = v;
        });

        if (init) {
            InitDefaultsCommon();
        }
        return this;
    };
    
    function Tokenizer(graph, input, punctuation) {
        punctuation = punctuation || ",;.";
        
        var position = 0;
        var data = input;

        function PeekChar(forward) {
            forward = forward || 0;
            return position + forward >= data.length ? undefined : data[position + forward];
        }

        function EatChar() {
            return position >= data.length ? undefined : data[position++];
        }

        function CreateWhitespaces() {
            var start = position;
            while (true) {
                var ch = PeekChar();
                if (ch && " \r\n\t"._contains(ch)) {
                    EatChar();
                } else {
                    break;
                }
            }
            if (start != position) {
                return { type: "Space", value: data.substring(start, position), skip: true };
            }
        }

        function CreateComment() {
            var start = position;
            if (PeekChar() == "#") {
                while (true) {
                    var ch = EatChar();
                    if (!ch || ch === "\n") {
                        return { type: "Comment", value: data.substring(start, position), skip: true };
                    }
                }
            }
        }
        
        function IsCharURI(ch) {
            return "%^:_~.-!$&'*+;=/?#@()"._contains(ch) || ch.match(/[0-9a-z]/i);
        }

        function CreateNumber() {
            var start = position;
            if ("+-0123456789"._contains(PeekChar())) {
                while ("+-.0123456789eE"._contains(PeekChar())) {
                    EatChar();
                }
                return { type: "Number", value: data.substring(start, position) };
            }
        }

        function CreateSymbol() {
            var start = position;
            while (IsCharURI(PeekChar())) {
                EatChar();
            }
            if (start != position) {
                return { type: "Symbol", value: data.substring(start, position) };
            }
        }

        function CreatePunctuation() {
            var start = position;
            var ch = PeekChar();
            if (!ch) {
                return { type: "EOF" };
            }
            if (punctuation._contains(ch)) {
                EatChar();
                return { type: "Punctuation", value: data.substring(start, position) };
            }
        }
        
        function PeekIfTripleQuote(quote, ch) {
            return quote === ch &&
                   quote === PeekChar() &&
                   quote === PeekChar(1);
        }

        function CreateString() {
            var start = position;
            if (PeekChar() === '"') {
                var esc = false;
                var quote = EatChar();

                // Check if longString (http://www.w3.org/TeamSubmission/turtle/#longString)
                var tripleQuote = PeekIfTripleQuote(quote, quote);
                if (tripleQuote) {
                    EatChar();
                    EatChar();
                }

                while (true) {
                    var ch = EatChar();
                    graph.Assert(ch, "ParseError");
                    
                    if (!esc && ch === quote) {
                        if (tripleQuote && PeekIfTripleQuote(quote, ch)) {
                            EatChar();
                            EatChar();
                            break;
                        } else if (!tripleQuote) {
                            break;
                        }
                    }
                    esc = ch === "\\";
                }

                // Pick up trailing language tag or type specification
                while (true) {
                    var ch = PeekChar();
                    if (!IsCharURI(ch) && ch != "<" && ch != ">") {
                        break;
                    }
                    EatChar();
                }
                return { type: "String", value: data.substring(start, position) };
            }
        }

        function CreateUri() {
            var start = position;
            if (PeekChar() === "<") {
                while (true) {
                    var ch = EatChar();
                    graph.Assert(ch, "ParseError");
                    if (ch === ">") {
                        break;
                    }
                }
                return { type: "Uri", value: data.substring(start, position) };
            }
        }

        this.GetNextToken = function () {
            while (true) {
                var token = CreateWhitespaces() ||
                            CreateComment() ||
                            CreatePunctuation() ||
                            CreateString() ||
                            CreateUri() || 
                            CreateNumber() ||
                            CreateSymbol();
                if (!token.skip) {
                    return token;
                }
            }
        }

        return this;
    }

    //      _____                       _
    //     | ____|_  ___ __   ___  _ __| |_ ___
    //     |  _| \ \/ / '_ \ / _ \| '__| __/ __|
    //     | |___ >  <| |_) | (_) | |  | |_\__ \
    //     |_____/_/\_\ .__/ \___/|_|   \__|___/
    //                |_|
   
    module.exports.MapNamespaces = MapNamespaces;
    module.exports.Tokenizer = Tokenizer;

// DO NOT PLACE ANYTHING BELOW THE FOLLOWING LINE -- IT
// BRACKETS THE ENTIRE FILE.
})();
