// ----------------------------------------------------------------------------------------------------------------------
//   Copyright 2014 (c) Microsoft Corporation.
//   Summary: Defines global prototype definitions.
// ----------------------------------------------------------------------------------------------------------------------

//  _   _   _ _   _ _ _ _   _
// ( ) | | | | |_(_) (_) |_(_) ___  ___
//  \| | | | | __| | | | __| |/ _ \/ __|
//     | |_| | |_| | | | |_| |  __/\__ \
//      \___/ \__|_|_|_|\__|_|\___||___/

(function () {
    "use strict"; // whole file

    //   ___  _     _           _
    //  / _ \| |__ (_) ___  ___| |_
    // | | | | '_ \| |/ _ \/ __| __|
    // | |_| | |_) | |  __/ (__| |_
    //  \___/|_.__// |\___|\___|\__|
    //           |__/

    // _extends
    //  Wraps Object.defineProperty to extend the current object.
    //  We always set "enumerable=false" because we don't want
    //  the new property (on prototype usually) on _forEach.
    //
    //  Any functions we are adding have _ prefix to distinguish
    //  from standard functions.
    //
    //  Add get/set accessors to the object:
    //    Object.defineProperty(obj, "newAccessorProperty", {
    //        set: function (x) {
    //            this.newaccpropvalue = x;
    //        },
    //        get: function () {
    //            return this.newaccpropvalue;
    //        },
    //        enumerable: false
    //    });
    // Add a data property to the object.
    //    Object.defineProperty(obj, "newDataProperty", {
    //        value: 101,
    //        writable: true,
    //        enumerable: true,
    //        configurable: true
    //    });

    var _extends = function (nym, fn) {
        var defn = { enumerable: false };

        if (typeof (fn) === "function") {
            defn.value = fn;
        } else if (fn instanceof (Object)) {
            if (!fn.enumerable) {
                fn.enumerable = false;
            }
            defn = fn;
        } else {
            throw "Object._extend: Illegal arg " + fn;
        }
        
        // TODO: Should we check if we are stomping?
        Object.defineProperty(this, nym, defn);
    };

    // Use _extends to inject the definition itself!
    // This is used (instead of Object.prototype._extends assignment)
    // so that enumerable is set to false as well.
    // And validate by dogfooding.
    _extends.call(Object.prototype, "_extends", _extends);
    
    Object.prototype._extends("_property", function (name) {
        var value = Object.getOwnPropertyDescriptor(this, name);
        if (value) {
            return value;
        }
        var parent = this._parent;
        if (parent) {
            return parent._property(name);
        }
    });
    
    //
    // create child object
    //
    Object.prototype._extends("_createChild", function () {
        var obj = Object.create(this);
        if (this._initChild) {
            this._initChild.apply(obj, arguments);
        }
        return obj;
    });

    Object.prototype._extends("_parent", {
        get: function () { return Object.getPrototypeOf(this); }
    });

    //
    // Keys
    //  Returns keys that are enumerable and hasOwnProperty
    Object.prototype._extends("_keys", {
        get: function () { return Object.keys(this); }
    });

    //
    // _forEach iterator
    //  If you use "this" within the function, then wrap the call as:
    //      someObject._forEach( this._thunk( function(k,v) { this.
    Object.prototype._extends("_forEach", function (fn) {
        var THIS = this;
        this._keys._forEach(function (k, v) {
            return fn.call(THIS, v, THIS[v]);
        });
    });
    
    Object.prototype._extends("_map", function (fn) {
        var THIS = this;
        var ret = [];
        this._keys._forEach(function (k, v) {
            ret.push(fn.call(THIS, v, THIS[v]));
        });
        return ret;
    });
    
    Object.prototype._extends("_any", function (fn) {
        var THIS = this;
        return this._keys._any(function (k, v) {
            return fn.call(THIS, v, THIS[v]);
        });
    });

    Object.prototype._extends("_all", function (fn) {
        var THIS = this;
        return this._keys._all(function (k, v) {
            return fn.call(THIS, v, THIS[v]);
        });
    });

    Object.prototype._extends("_allZip", function (that, fn) {
        var keys = this._keys;
        if (!(that instanceof (Object)) ||
            that._keys.length !== keys.length) {
            return false;
        }
        var THIS = this;
        return keys._all(function (k, v) {
            return (v in that) && fn.call(THIS, v, THIS[v], that[v]);
        });
    });

    //
    // _equals
    //  Check if the two objects are the same
    Object.prototype._extends("_equals", function (obj1, obj2) {
        // footwork to push this into obj1
        switch (arguments.length) {
            case 1: return this._equals(this, obj1);
            case 2: break;
            default:
            case 0: throw "Object._equals: Illegal arg";
        }

        if (obj1 === obj2) {
            return true;
        }

        if (obj1 instanceof (Object)) {
            return obj1._allZip(obj2, function (k, v1, v2) { return Object._equals(v1, v2); });
        }

        return false;
    });

    // _merge
    //  Merges two objects and returns a new object
    //  If there are conflicting properties, returns undefined
    Object.prototype._extends("_merge", function (obj2) {
        if (!obj2) {
            return;
        }

        var ret = this instanceof (Array) ?
                        [] :
                        this._parent._createChild();

        // Shallow copy properties
        this._forEach(function (k, v) {
            ret[k] = v;
        });

        // Shallow copy properties if no conflict
        return obj2._all(function (k, v) {
            var v2 = ret[k];
            if (!v2) {
                ret[k] = v;
                return true;
            } else {
                return v2._equals(v);
            }
        }) ? ret : undefined;
    });

    Object.prototype._extends("_subsetKeys", function (obj2) {
        return this._all(function (k, v) { return obj2.hasOwnProperty(k); });
    });
    
    Object.prototype._extends("_clear", function () {
        this._all(function (k, v) { return delete this[k]; });
    });

    //
    // _thunk
    //  Replaces "this" pointer with the thunks.
    //  This is needed because obj._forEach, obj._all, etc.
    //  sends in "obj" and not usually the context you want.
    //
    //  e.g. without thunk, this===arguments
    //       with thunk, this===logger
    //  var logger = {
    //      toString: function () {
    //          arguments._forEach( this._thunk(
    //              function (n, v) { return this; }
    //      ))}
    //  };
    //
    //  Squirrel away the "this" pointer for applying
    //  within the fn.
    Object.prototype._extends("_thunk", function (fn) {
        var THIS = this;
        return function () { return fn.apply(THIS, arguments); };
    });

    //
    // _clone with transformation
    //  Make deep copy of the objects.
    //  Transform function is applied if provided
    Object.prototype._extends("_clone", function (transform) {
        if (transform !== undefined) {
            var obj = transform(this);
            if (obj) {
                return obj;
            }
            // Prevents replacement containing pattern and repeating
        }

        if (!this._suppressClone) {
            // Obtain the prototype of the original object and create
            // a new object by using that prototype.
            // For Array, use a new Array (or else you get an object)
            var ret = this instanceof (Array) ?
                        [] :
                        this._parent._createChild();

            // Copy the properties of the original object to the new object.
            this._forEach(function (k, v) {
                ret[k] = v instanceof (Object) ? v._clone(transform) : v;
            });

            return ret;
        }
        return this;
    });

    Object.prototype._extends("_exists", function (matcher) {
        return matcher(this) ||
            (!this._suppressClone &&
             this._any(function (k, v) { return v instanceof (Object) && v._exists(matcher); }));
    });


    //     _
    //    / \   _ __ _ __ __ _ _   _
    //   / _ \ | '__| '__/ _` | | | |
    //  / ___ \| |  | | | (_| | |_| |
    // /_/   \_\_|  |_|  \__,_|\__, |
    //                         |___/

    //
    // _forEach iterator
    //  If you use "this" within the function, then wrap the call as:
    //      someObject._forEach( this._thunk( function(k,v) { this.
    Array.prototype._extends("_forEach", function (fn) {
        var l = this.length;
        for (var key = 0; key < l; key++) {
            if (key in this) {
                fn.call(this, key, this[key]);
            }
        }
    });

    Array.prototype._extends("_any", function (fn) {
        var l = this.length;
        for (var key = 0; key < l; key++) {
            if (key in this) {
                var ret = fn.call(this, key, this[key]);
                if (ret) {
                    return ret;
                }
            }
        }
    });

    Array.prototype._extends("_all", function (fn) {
        var l = this.length;
        for (var key = 0; key < l; key++) {
            if (key in this) {
                if (!fn.call(this, key, this[key])) {
                    return false;
                }
            }
        }
        return true;
    });

    Array.prototype._extends("_allZip", function (that, fn) {
        var l = this.length;
        if (!(that instanceof (Array)) ||
            that.length !== l) {
            return false;
        }
        for (var key = 0; key < l; key++) {
            if (key in this) {
                if (!(key in that) || !fn.call(this, key, this[key], that[key])) {
                    return false;
                }
            }
        }
        return true;
    });
    
    Array.prototype._extends("_clear", function (that, fn) {
        this.splice(0, this.length);
    });

    Array.prototype._extends("_replace", function (from, len, replace) {
        len = len || 1;
        replace = replace || [];
        return this.slice(0, from).concat(replace).concat(this.slice(from + len));
    });

    Array.prototype._extends("_pushUnique", function (d) {
        if (!this._any(function (k, v) { return v._equals(d); })) {
            this.push(d);
        }
    });

    //  ____  _        _
    // / ___|| |_ _ __(_)_ __   __ _
    // \___ \| __| '__| | '_ \ / _` |
    //  ___) | |_| |  | | | | | (_| |
    // |____/ \__|_|  |_|_| |_|\__, |
    //                         |___/

    // _format
    //  e.g. "Test{0}"._format("-pass") -> "Test-pass"
    String.prototype._extends("_format", function () {
        var ret = this;
        for (var i = 0; i < arguments.length; i++) {
            var repl = arguments[i];
            if (repl.toString) {
                repl = repl.toString();
            }
            ret = ret._replace(new RegExp("\\{" + i + "\\}", "gm"), repl);
        }
        return ret;
    });

    String.prototype._extends("_replace", function (find, repl) {
        // Escape $ to $$ in the replacement string to avoid getting
        // trapped by special patterns ($$, $&, $`, $', $n, $nn)
        return this.replace(find, repl.replace(/\$/g, "$$$$"));
    });

    // _repeat
    //  similar to spaces function
    String.prototype._extends("_repeat", function (n) {
        return new Array((n || 0) + 1).join(this);
    });

    String.prototype._extends("_htmlify", function () {
        return this.replace(/ /g, "&nbsp;")
                   .replace(/\n/g, "<br/>");
    });

    String.prototype._extends("_startsWith", function (str) {
        return this.indexOf(str) === 0;
    });

    String.prototype._extends("_endsWith", function (str) {
        return this.indexOf(str, this.length - str.length) !== -1;
    });

    String.prototype._extends("_contains", function (str) {
        return this.indexOf(str) !== -1;
    });

    String.prototype._extends("_find", function (str) {
        var index = this.indexOf(str);
        return index !== -1 ? this.substring(index) : "";
    });

    String.prototype._extends("_findLast", function (str) {
        var index = this.lastIndexOf(str);
        return index !== -1 ? this.substring(index) : "";
    });
    
    // http://stackoverflow.com/questions/105034/how-to-create-a-guid-uuid-in-javascript
    String._extends("_createGuid", function () {
        function s4() {
            return Math.floor((1 + Math.random()) * 0x10000)
               .toString(16)
               .substring(1);
        }
        return s4() + s4() + '-' + s4() + '-' + s4() + '-' +
           s4() + '-' + s4() + s4() + s4();
    });

    // DO NOT PLACE ANYTHING BELOW THE FOLLOWING LINE -- IT
    // BRACKETS THE ENTIRE FILE.
})();
