// ----------------------------------------------------------------------------------------------------------------------
//   Copyright 2014 (c) Microsoft Corporation.
//   Summary: Defines the Webserver for internal testing. INTERNAL USE ONLY.
// ----------------------------------------------------------------------------------------------------------------------
(function () {
    "use strict";  // whole file

    //  ___                            _
    // |_ _|_ __ ___  _ __   ___  _ __| |_ ___
    //  | || '_ ` _ \| '_ \ / _ \| '__| __/ __|
    //  | || | | | | | |_) | (_) | |  | |_\__ \
    // |___|_| |_| |_| .__/ \___/|_|   \__|___/
    //               |_|

    var graph = require("./Graph").CreateGraph();

    // Express http://expressjs.com/guide.html
    var express = require("express");
    var path = require("path");
    var fs = require("fs");
    var childProc = require("child_process");
    
    // Create generated folder for inlined files.
    var cacheDir = "Generated";
    fs.stat(cacheDir, function (err, stats) {
        if (err) {
            fs.mkdir(cacheDir, function () { });
        }
    });

    // __        __   _    ____
    // \ \      / /__| |__/ ___|  ___ _ ____   _____ _ __
    //  \ \ /\ / / _ \ '_ \___ \ / _ \ '__\ \ / / _ \ '__|
    //   \ V  V /  __/ |_) |__) |  __/ |   \ V /  __/ |
    //    \_/\_/ \___|_.__/____/ \___|_|    \_/ \___|_|

    var webServer = {
        app: express(),

        graph: graph,

        log: function () {
            var args = Array.prototype.slice.call(arguments);
            if (args.length > 1) {
                args[0] = { "#magenta": args[0] };
            }
            this.graph.Log.apply(undefined, args);
        },

        streamToString: function (stream, completion) {
            var data = "";
            stream.on("data", function (newData) {
                data += newData;
            });

            stream.on("end", function () {
                completion(data);
            });
        },

        collection: null,
        reqNum: 0
    };

    // loadFile
    //  Loads the specified file and calls the callback with the contents.
    //  If there is a failure, callback is not called.
    webServer.loadFile = function (f, callback) {
        var filename = path.join(module.filename, "..", f);
        if (fs.existsSync(filename)) {
            var data = fs.readFileSync(filename, "utf-8");
            callback(data);
        } else {
            callback(undefined);
        }
    };
    
    // saveFile
    webServer.saveFile = function (f, data) {
        fs.writeFileSync(f, data, "utf-8");
    };
    
    // compile
    //  Inline all dependent modules into a single file.  Base code running in the
    //  browser should not refer to any of the nodeJS classes and they are not
    //  inlined during the compilation.
    webServer.compile = function (f, callback, modules) {
        modules = modules || {};

        // Keep this in sync with Browser.js::require.modules._define
        var baseName = f.split("/").slice(-1)[0].toLowerCase().replace(/\.js$/g, "");
        if (baseName in modules) {
            return "";
        }
        modules[baseName] = true;

        var source;
        this.loadFile(baseName + ".js", function (data) { source = data; });
        if (!source) {
            return "";
        }

        var regexp = /require\s*\(\s*(["'])(.*?)\1\s*\)/g;

        var includes = '';
        var matches = source.match(regexp) || [];
        matches._forEach(function (k, v) {
            includes += webServer.compile(v.replace(regexp, '$2'), undefined, modules);
        });

        var header = includes +
                     "module = require.modules._define(\"{0}\");\n"._format(baseName) +
                     "exports = module.exports;\n";

        if (source.match(/\/\/ INLINEMODULES/g)) {
            // Top level module
            // need to escape $ in replacement because they have special meaning in regexp
            var data = source._replace(/\/\/ INLINEMODULES/g, header);
            callback(data);
            return data;
        } else {
            return header + source + "\n";
        }
    };
    
    webServer.minJS = function (inputFile, callback) {
        var command = '"External\\AjaxMin\\AjaxMin.exe" "{0}"'._format(inputFile);
        // '"c:\\Program Files (x86)\\Microsoft\\Microsoft Ajax Minifier\\AjaxMin.exe" "{0}"'._format(inputFile);
        childProc.exec(command, function (error, stdout, stderr) {
            callback(stdout);
        });
    };

    // Experimental web-page server
    webServer.indexPageServer = function (req, res) {
        var file = path.join(module.filename, "../index.html");
        webServer.log("HTTP`Send", path.basename(file));
        res.statusCode = 200;
        res.setHeader("Content-Type", "text/html");
        res.sendFile(file);
    };
    
    webServer.javaScript = function (res, filename, compile, compress) {
        var ext = path.extname(filename);

        if (compile) {
            return this.compile(filename, this._thunk(function (payload) {
                var outputFile = path.join(cacheDir, filename._replace(ext, ".Debug" + ext));
                this.saveFile(outputFile, payload);
                return this.javaScript(res, outputFile, undefined, compress);
            }));
        }
        if (compress) {
            return this.minJS(filename, this._thunk(function (payload) {
                var outputFile = filename._replace(".Debug" + ext, ext);
                this.saveFile(outputFile, payload);
                return this.javaScript(res, outputFile);
            }));
        }

        webServer.log("HTTP`Send", path.basename(filename));
        webServer.loadFile(filename, function (payload) {
            res.statusCode = 200;
            res.setHeader("Content-Type", "text/javascript");
            res.end(payload);
        });
    };

    webServer.getServer = function (req, res) {
        webServer.log("HTTP`Get");
        var callback = function (expr) { webServer.postReply(expr, req, res); };
        webServer.graph.Context.SetCallback(callback);
        webServer.graph.Evaluate(req.query.InputForm, "InputForm");
    };

    webServer.postServer = function (req, res) {
        webServer.reqNum++;
        webServer.streamToString(req, function (str) {
            // log("RECEIVED POST STR: \"" + str + "\"");
            var form;
            if (req.is("application/json")) {
                form = "JaqSONForm";
            } else if (req.is("text/plain")) {
                form = "InputForm";
            }

            var callback = function (expr) { webServer.postReply(expr, req, res); };
            webServer.graph.Context.SetCallback(callback);
            webServer.graph.Evaluate(str, form);
        });
    };

    webServer.postReply = function (expr, req, res) {
        var graph = expr.graph;

        webServer.log("HTTP`In[" + webServer.reqNum + "]:=", graph.Context.Input.toInputForm());
        webServer.log("HTTP`Out[" + webServer.reqNum + "]=", { "#bold": expr });

        if (webServer.collection) {
            webServer.collection
                .observize()
                .insert(graph.Context.Input.toJSON())
                .completion(function () {
                });
        }

        // Look at accept headers
        res.statusCode = 200;
        res.setHeader("Access-Control-Allow-Origin", "*"); // This allows other sites to call
        if (req.accepts("text/plain")) {
            res.setHeader("Content-Type", "text/plain");
            res.end(expr.toString());
        } else if (req.accepts("application/json")) {
            res.setHeader("Content-Type", "application/json");
            res.end(expr.toJaqSONForm());
        } else {
            res.setHeader("Content-Type", "text/plain");
            res.end(expr.toString());
        }
    };

    webServer.start = function () {
        this.log({ "#green": "(* ClientGraph bootup *)" });
        this.app.get("/", this.indexPageServer);
        this.app.get("/ClientGraph.Debug.js", this._thunk(function (req, res) { return this.javaScript(res, "ClientGraph.js", true); }));
        this.app.get("/ClientGraph.js", this._thunk(function (req, res) { return this.javaScript(res, "ClientGraph.js", true, true); }));
        this.app.get("/ClientGraphTest.Debug.js", this._thunk(function (req, res) { return this.javaScript(res, "ClientGraphTest.js", true); }));
        this.app.get("/ClientGraphTest.js", this._thunk(function (req, res) { return this.javaScript(res, "ClientGraphTest.js", true, true); }));
        this.app.listen(5000); //process.env.PORT || 5000);
    };

    //  _____                       _
    // | ____|_  ___ __   ___  _ __| |_ ___
    // |  _| \ \/ / '_ \ / _ \| '__| __/ __|
    // | |___ >  <| |_) | (_) | |  | |_\__ \
    // |_____/_/\_\ .__/ \___/|_|   \__|___/
    //            |_|

    module.exports.webServer = webServer;

    // DO NOT PLACE ANYTHING BELOW THE FOLLOWING LINE -- IT
    // BRACKETS THE ENTIRE FILE.
})();
