﻿var answers = require('../lib/answers.js');
var answersMocks = require('./answersMocks.js');
var httpMocks = require('./httpMocks.js');

var mockProactive;
var mockRouter;
var req;


exports.getProactiveTests = {
  setUp: function (callback) {
    this.mockProactive = new answersMocks.MockProactive();
    this.mockRouter = new answersMocks.MockRouter();
    this.req = httpMocks.createMockRequest();
    callback();
  },
  basePageReturned: function (test) {
    test.expect(1);
    
    answers.start(this.mockRouter, this.mockProactive);
    
    this.mockProactive.setGetProactiveBehavior('noError');
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(200, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.proactiveGetHandler(this.req, resp);
  },
  noConnectionReturnError: function (test) {
    test.expect(1);
    
    this.mockProactive.setGetProactiveBehavior('noConnection');
    
    answers.start(this.mockRouter, this.mockProactive);
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(503, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.proactiveGetHandler(this.req, resp);
  },
  timeoutError : function (test) {
    test.expect(1);
    
    this.mockProactive.setGetProactiveBehavior('requestTimeout');
    
    answers.start(this.mockRouter, this.mockProactive);
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(504, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.proactiveGetHandler(this.req, resp);
  },
  genericError : function (test) {
    test.expect(1);
    
    this.mockProactive.setGetProactiveBehavior('genericError');
    
    answers.start(this.mockRouter, this.mockProactive);
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(500, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.proactiveGetHandler(this.req, resp);
  }
};

exports.getAnswersTests = {
  setUp: function (callback) {
    this.mockProactive = new answersMocks.MockProactive();
    this.mockRouter = new answersMocks.MockRouter();
    this.req = httpMocks.createMockRequest();
    answers.start(this.mockRouter, this.mockProactive);
    
    callback();
  },
  getAnswersEmptyHeaders: function (test) {
    test.expect(1);
    
    // assign empty list
    this.req.headers = [];
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(500, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.answersGetHandler(this.req, resp);
  },
  getAnswersNullHeaders: function (test) {
    test.expect(1);
    
    // null headers
    this.req.headers = null;
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(500, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.answersGetHandler(this.req, resp);
  },
  getAnswersUndefinedHeaders: function (test) {
    test.expect(1);
    
    // undefined headers
    this.req.headers = undefined;
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(500, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.answersGetHandler(this.req, resp);
  },
  getAnswersWrongAcceptHeaders: function (test) {
    test.expect(1);
    
    // assign unexpected header
    this.req.headers['accept'] = 'text/html';
    
    var resp = httpMocks.createMockServerResponse();
    
    resp.on('finish', function () {
      test.equal(500, resp.getStatusCode());
      test.done();
    });
    
    this.mockRouter.routeHandlers.answersGetHandler(this.req, resp);
  },
  getAnswersSuccess: function (test) {
    test.expect(3);
    
    this.req.headers['accept'] = 'application/json';
    
    callGetAnswersAndVerify('noError', this.req, this.mockProactive, this.mockRouter, function (resp) {
      var jsonData = JSON.parse(resp.getContent());
      var expectedJson = answersMocks.getMockAnswersSync();
      
      test.equal(resp.getContentType(), 'application/json');
      test.equal(resp.getStatusCode(), 200);
      test.deepEqual(jsonData, expectedJson);
      test.done();
    });
  },
  getAnswersNullErrorAndNullResults: function (test) {
    test.expect(2);
    
    this.req.headers['accept'] = 'application/json';
    
    callGetAnswersAndVerify('invalidState', this.req, this.mockProactive, this.mockRouter, function (resp) {
      test.equal(resp.getContentType(), 'application/json');
      test.equal(resp.getStatusCode(), 200);
      test.done();
    });
  },
  getAnswersGenericError: function (test) {
    test.expect(2);
    
    // In this case, the previous call to proactive had propagated an error
    this.req.headers['accept'] = 'application/json';
    
    callGetAnswersAndVerify('genericError', this.req, this.mockProactive, this.mockRouter, function (resp) {
      test.equal(resp.getContentType(), 'text/html');
      test.equal(resp.getStatusCode(), 500);
      test.done();
    });
  }
};

var callGetAnswersAndVerify = function (proactiveBehaviorName, mockRequest, mockProactive, mockRouter, verify) {
  var proactiveResponse = httpMocks.createMockServerResponse();
  mockProactive.setGetProactiveBehavior(proactiveBehaviorName);
  
  proactiveResponse.on('finish', function () {
    console.log('Call to proactive finished. Status Code: %s', proactiveResponse.getStatusCode());
  });
  
  mockRouter.routeHandlers.proactiveGetHandler(mockRequest, proactiveResponse);
  
  var resp = httpMocks.createMockServerResponse();
  
  resp.on('finish', function () {
    verify(resp);
  });
  
  mockRouter.routeHandlers.answersGetHandler(mockRequest, resp);
};