var util = require('util');
var isAndroid = (process.platform == 'android');
var javaEventsManager = require('./javaEventsManager.js');
var utils = require('./utils.js');
var logger = require('logger')(__filename);

(function initConsoleLog() {

    if(!isAndroid) return;
    var androidLog = require('androidlog');
    var logLimit = 1023;

    function logMessage(verbosity, msg) {
        var length = msg.length;
        for(var i=0; i<length; i+=logLimit)
        {
            if(i+logLimit < length)
                androidLog[verbosity](msg.substring(i, i+logLimit));
            else
                androidLog[verbosity](msg.substring(i, length));
        }
    }

    console.log =       function() { logMessage('verbose',  util.format.apply(null, arguments)); }
    console.verbose =   function() { logMessage('verbose',  util.format.apply(null, arguments)); }
    console.debug =     function() { logMessage('debug',    util.format.apply(null, arguments)); }
    console.warn =      function() { logMessage('warn',     util.format.apply(null, arguments)); }
    console.info =      function() { logMessage('info',     util.format.apply(null, arguments)); }
    console.error =     function() { logMessage('error',    util.format.apply(null, arguments)); }
})();

var java = isAndroid ? require('nodejavabridge_bindings') : null;
var container = isAndroid ? java.callStaticMethodSync('com.microsoft.bing.dss.platform.infra.Container', 'getInstance') : null;

logger.log('node is running on threadId', utils.getThreadId());

function notifyNodeReady(port, authCookieUuid) {
    logger.log('notifying cortana app that node is ready, with port: ', port);
    if(!port) {
        throw new Error('port was not provided');
    }

    java.callStaticMethodSync('com.microsoft.bing.dss.nodelib.NodeReadyEventSource', 'notifyNodeReady', port, authCookieUuid);
}

function javaCallbackWrapper(cb, title) {
    title = title || 'untitled';
    var runnableProxy = createJavaProxy('java.lang.Runnable', {
      run: function () {
        var arguments = nodeJavaCb.getParams();
        cb.apply(null, arguments);
      }
    });

    var nodeJavaCb = java.newInstanceSync('com.microsoft.bing.dss.nodelib.NodeJavaCallback', runnableProxy, title);
    return nodeJavaCb;
}

function getDevice() {

    if(!isAndroid) return {};

    var components = {},
        device = {};

    var componentsName = [
        'notifications',
        'applications',
        'scheduler',
        'screen',
        'telephony',
        'regions',
        'battery',
        'network',
        'signals',
        'location',
        'modeOfTransport',
        'localStorage',
        'sharedStorage',
        'messaging',
        'httpClient',
        'bluetooth',
        'audio',
        'media' ,
        'calendar',
        'contacts',
        'proximity',
        'position',
        'nfc',
        'gestures',
        'flashlight',
        'microphone',
        'shortcuts',
        'attentionmanager',
        'timers',
        'microphone',
        'shortcuts',
        'serviceManager',
        'speech',
        'textSpeech',
        'reminder',
        'authentication',
        'utils',
        'db',
        'uploadPolicyManager',
        'sample',
        'headers',
        'locationApi',
        'appUtils'
    ];

    function getComponent(componentName) {

        var component = components[componentName];
        if(component)
            return component;

        component = container.getComponent(componentName);
        var wrapper = {};

        for(var method in component) {

            var origMethodObj = component[method];

            (function(componentName, method, origMethodObj) {

                wrapper[method] = function() {
                    var args = [].slice.call(arguments);
                    for(var i=0; i<args.length; i++) {
                        if(typeof args[i] === 'function') {
                            //this is a function- wrap it with a node-java proxy
                            var title = componentName + '.' + method;
                            args[i] = javaCallbackWrapper(args[i], title);
                        }
                    }
                    return origMethodObj.apply(component, args);
                }
            })(componentName, method, origMethodObj);
        }
        components[componentName] = wrapper;
        return wrapper;
    }

    function createGetter(componentName) {
        device.__defineGetter__(componentName, function() {
            return getComponent(componentName);
        });
    }

    for(var i=0; i<componentsName.length; i++) {
        createGetter(componentsName[i]);
    }

    return device;
}

function getAuthenticationTokens(cb) {
  var cbImp = createJavaProxy('com.microsoft.bing.dss.authlib.TokensIssuedCallback', {
    onCompleted: function (result) {
      return cb(result.getException(), {
            rpsToken: result.getRpsToken(),
            aisToken: result.getAisToken(),
            muid: result.getMuid()
        });
    }
  });

  device.authentication.getTokens(cbImp);
}

function getConfiguration() {
  return container.getConfigurationManager();
}

function createJavaInstance() {
    logger.log('creating a Java instance with the following arguments', arguments);
    return java.newInstanceSync.apply(java, arguments);
}

function createJavaProxy() {
    logger.log('creating a Java proxy with the following arguments', arguments);
    return java.newProxy.apply(java, arguments);
}

function createJavaArray() {
  logger.log('creating a Java array with the following arguments', arguments);
  return java.newArray.apply(java, arguments);
}

var device = getDevice();

var api = {
     notifyNodeReady: notifyNodeReady,
     device: device,
     // used to create instances of Java classes
     createJavaInstance: createJavaInstance,
     // used to create implementations for Java interfaces
     createJavaProxy: createJavaProxy,
     // used to create instances of Java primitive arrays,
     createJavaArray: createJavaArray,
     // used to wrap javascript callbacks before handing it over to a Java API
     javaCallbackWrapper: javaCallbackWrapper,
     getAuthenticationTokens: getAuthenticationTokens,
     getConfiguration: getConfiguration,
     utils: utils,
     eventsManager: javaEventsManager(java, javaCallbackWrapper),
     isAndroid: isAndroid
};

for(var obj in api) {
    var member = api[obj];
    if(typeof member === 'function') {
        (function(obj, origMethodObj) {
            api[obj] = function() {
                if(!isAndroid) return logger.warn('this is not android, cannot call %s() function', obj);
                logger.log('java bridge API called:', obj);
                return origMethodObj.apply(api, arguments);
            }
        })(obj, member);
    }
}

module.exports = api;
