﻿// Copyright (c) Microsoft Corporation.  All rights reserved.
/// <disable>JS2043.RemoveDebugCode</disable>

var middler = require('middler');
var path = require('path');
var templateUtils = require('template-utils');
var send = require('send');
var collector = require('metrics-collector');
var router = middler();
var staticPath = path.join(__dirname, 'static');

var HIST_BINS_NUM = 6;
var HIST_FIRST_BIN_UPPER_BOUND = 125;
SESSIONS_NUM_TO_DISPLAY = 20;

function onChartDataReady(pageMetadata, err) {
  if (err) {
    // save the last error to be used when calling the final callback
    pageMetadata.lastError = err;
  }

  // check if all charts data is ready
  if (++pageMetadata.readyChartsCounter == pageMetadata.totalChartsNum) {
    if (pageMetadata.lastError) {
      return pageMetadata.finalCb(pageMetadata.lastError);
    }
    
    // get the card's template
    var templatePath = path.join(__dirname, 'templates', 'metricsCardHtmlOnly.dt');
    templateUtils.loadTemplate(templatePath, function (err, template) {
      if (err) {
        return pageMetadata.finalCb(err);
      }
      // instantiate the template with the charts data
      var cardContent = template(pageMetadata.pageData);

      var resultObject = {
        Content: {
          HTML: cardContent
        }
      };

      return pageMetadata.finalCb(null, resultObject);

    });
  }
};



function appendRequestsDurationHistogramData(chartIndex, pageMetadata, chartDataReadyCb) {
  var chartObj = {headline: 'Request durations distribution', subHeadline: 'Over last ' + collector.LIMIT + ' requests'};

  collector.getRequestsTimestampBounds(function(err, rows) {
    if (err) {
      return onChartDataReady(err);
    }
    
    if (rows.length == 0) {
      return onChartDataReady(chartObj);
    }

    function handleHistogramValues (err, vals) {
      if (err) {
        return onChartDataReady(err);
      }

      if (vals.length == 0) {
        return onChartDataReady(chartObj);
      }
        
      var xLabel = 'Request duration in ms';
      var yLabel = '# Requests';
  
      // create arrays to arrange axes values
      var xAxis = new Array(vals.length + 1);
      var yAxis = new Array(vals.length + 1);
      xAxis[0] = xLabel;
      yAxis[0] = yLabel;
  
      for (var i = 0; i < vals.length; ++i) {
        xAxis[i + 1] = vals[i].binLowerBound + (vals[i].binUpperBound ? (' - ' + vals[i].binUpperBound) : ' <');
        yAxis[i + 1] = vals[i].val;
      }

      // arrange chart data in the proper format for c3.js module
      var chartData = {
        type: 'bar',
        labels: true,
        columns: [xAxis, yAxis],
        x: xLabel,
        y: yLabel
      };
      
      var axisConf = {
        y: {
          label: {
            text: yLabel,
            position: 'outer-middle'
          }
        },
        x: {
          type: 'category',
          label: {
            text: xLabel,
            position: 'outer-center'
          },
          tick: {
            rotate: -45,
            multiline: false
          },
          height: 70
        }
      };
    
      chartObj.data = chartData;
      chartObj.axis = axisConf;

      pageMetadata.pageData.charts[chartIndex] = chartObj;
      chartDataReadyCb();
    };
  
    // get the histogram data from the database
    collector.getHistogramValues(rows[0].timestampLowerBound,
                                 rows[0].timestampUpperBound,
                                 HIST_BINS_NUM,
                                 HIST_FIRST_BIN_UPPER_BOUND,
                                 handleHistogramValues);
  });
};


function appendDomainToDurationMedianData (chartIndex, pageMetadata, chartDataReadyCb) {
  // get a mapping from domain to request duration median from the metrics database
  collector.getMedians(function (err, rows) {
    if (err) {
      return chartDataReadyCb(err);
    }

    // arrange axes values   
    var xLabel = 'Domains';
    var yLabel = 'Duration median';
    var xAxis = new Array(rows.length + 1);
    var yAxis = new Array(rows.length + 1);
    xAxis[0] = xLabel;
    yAxis[0] = yLabel;

    for (var i = 0; i < rows.length; ++i) {
      xAxis[i + 1] = rows[i].domain;
      yAxis[i + 1] = rows[i].median;
    }

    // arrange chart data in the proper format for c3.js module
    var chartData = {
      type: 'bar',
      labels: {
        format: function (v, id, i, j)  { return v; }
      },
      x: xLabel,
      y: yLabel,
      columns: [xAxis, yAxis]
    };

    var axisConf = {
      y: {
        label: {
          text: 'Median time in ms',
          position: 'outer-middle'
        },
        tick: {
          format: 'd3.format(",.0f")'
        }
      },
      x: {
        label: {
          text: xLabel,
          position: 'outer-center'
        },
        type: 'category',
        tick: {
          rotate: -45,
          multiline: false
        },
        height: 145
      }
    };

    pageMetadata.pageData.charts[chartIndex] = {data: chartData, axis: axisConf, headline: 'Median request duration', subHeadline: 'Over last ' + collector.LIMIT + ' requests'};
    chartDataReadyCb();
  });
}


function appendDomainToExceptionsNumData (chartIndex, pageMetadata, chartDataReadyCb) {
  collector.getDomainToExceptionsNum(function(err, rows) {
    if (err) {
      return chartDataReadyCb(err);
    }
    
    var chartObj = {headline: 'Exceptions number for domain', subHeadline: 'Over the last ' + collector.LIMIT + ' requests'};
    if (rows.length == 0) {
      pageMetadata.pageData.charts[chartIndex] = chartObj;
      return chartDataReadyCb();
    }
    
    var xLabel = 'Domains';
    var yLabel = '# Exceptions';
    xAxis = new Array(rows.length + 1);
    yAxis = new Array(rows.length + 1);
    xAxis[0] = xLabel;
    yAxis[0] = yLabel;
    for (var i = 0; i < rows.length; ++i) {
      xAxis[i + 1] = rows[i].domain;
      yAxis[i + 1] = rows[i].countVal;
    }

    var chartData = {
      type: 'bar',
      labels: true,
      x: xLabel,
      y: yLabel,
      columns: [xAxis, yAxis]
    };

    var axisConf = {
      y: {
        label: {
          text: yLabel,
          position: 'outer-middle'
        },
        tick: {
          format: 'd3.format(",.0f")'
        }
      },
      x: {
        label: {
          text: xLabel,
          position: 'outer-center'
        },
        type: 'category',
          tick: {
          rotate: -45,
          multiline: false
        },
        height: 145
      }
    };

    chartObj.data = chartData;
    chartObj.axis = axisConf;

    pageMetadata.pageData.charts[chartIndex] = chartObj;
    chartDataReadyCb();
  });
};


function appendDomainToMaximalRequestTimeData (chartIndex, pageMetadata, chartDataReadyCb) {
  collector.getSessionsTime(SESSIONS_NUM_TO_DISPLAY, function(err, rows) {
    if (err) {
      return chartDataReadyCb(err);
    }
    
    var chartObj = {headline: 'Session time', subHeadline: 'Measured during the last ' + SESSIONS_NUM_TO_DISPLAY + ' sessions'};

    if (rows.length == 0) {
      pageMetadata.pageData.charts[chartIndex] = chartObj;
      return chartDataReadyCb();
    }

    var yLabel = 'Session time in ms';

    var yAxis = new Array(rows.length + 1);
    yAxis[0] = yLabel;
    for (var i = 0; i < rows.length; ++i) {
      yAxis[i + 1] = parseInt(JSON.parse(rows[i].metadata).timeFromActivityStart);
    }

    var chartData = {
      labels: true,
      y: yLabel,
      columns: [yAxis]
    };

    var axisConf = {
      y: {
        label: {
          text: yLabel,
          position: 'outer-middle'
        },
        tick: {
          format: 'd3.format(",.0f")'
        }
      }
    };

    chartObj.data = chartData;
    chartObj.axis = axisConf;

    pageMetadata.pageData.charts[chartIndex] = chartObj;
    chartDataReadyCb();

  });
};


function appendMemoryConsumptionInLastSessionData (chartIndex, pageMetadata, chartDataReadyCb) {
  collector.allRowsDistinct('id', {type: collector.MEMORY, orderBy: 'timestamp', bDescending: true, limit: 2}, function(err, rows) {
    if (err) {
      return chartDataReadyCb(err);
    }
    
    var chartObj = {headline: 'Memory consumption', subHeadline: 'Measured during previous session'};

    if (rows.length == 0) {
      pageMetadata.pageData.charts[chartIndex] = chartObj;
      return chartDataReadyCb();
    }
    
    var yLabel = 'RSS value';
    var sessionId = (rows.length < 2) ? rows[0].id : rows[1].id;
    var memoryMeasurements = [yLabel];
    
    
    collector.eachRow({id: sessionId, type: collector.MEMORY, bASC: true},
      function(err, row) {
        if (err) {
          // skip this row
          return;
        }
        memoryMeasurements.push(row.intValue);
      },
      function(err, num) {
        if (err) {
          return chartDataReadyCb(err);
        }
        
        if (num > 0) {
          // prepare the c3 chart data
          var memoryChartData = {
            y: 'RSS value',
            columns: [memoryMeasurements]
          };

          var axisConf = {
            y: {
              label: {
                text: 'RSS value',
                position: 'outer-middle'
              },
              tick: {
                format: 'd3.format(",.0f")'
              }
            }
          };
          chartObj.data = memoryChartData;
          chartObj.axis = axisConf;
        }

        pageMetadata.pageData.charts[chartIndex] = chartObj;
        chartDataReadyCb();
      }
    );
  });
};  
  
var dataRetrievers = [
  appendRequestsDurationHistogramData,
  appendDomainToDurationMedianData,
  appendDomainToExceptionsNumData,
  appendDomainToMaximalRequestTimeData//,
  //appendMemoryConsumptionInLastSessionData temporary excluding chart
];
  

exports.produce = function (options, cb) {
  cb = cb || function () { };
  
  var pageMetadata = {
    lastError: null,
    readyChartsCounter: 0,
    totalChartsNum: 0,
    pageData: { basepath: exports.__basepath__ },
    finalCb: cb
  };
  
  pageMetadata.pageData.charts = new Array(dataRetrievers.length);
  
  var chartDataReadyCb = onChartDataReady.bind(undefined, pageMetadata);
  var chartsCounter = 0;
  
  dataRetrievers.forEach(function (func) {
    (function (chartIndex) {
      func.call(undefined, chartIndex, pageMetadata, chartDataReadyCb);
    })(chartsCounter++);
  });

  pageMetadata.totalChartsNum = dataRetrievers.length;
  
};

// any http request for the provider which is of the form /serviceName/* will be routed
// into our middler router by the answers service
exports.serve = router.handler.bind(router);


// catch the rest of the request starting with /seviceName and treat them as static file requests
router.get('*', function (req, res, next) {
  // req.params is set by the answers service which is calling into this provider
  var pathname = req.params[0];
  return send(req, pathname, { root: staticPath })
        .etag(false)
        .on('error', next)
        .pipe(res);
});
