﻿/// <disable>JS2043.RemoveDebugCode</disable>
/// <disable>JS2087.DoNotDeleteObjectProperties</disable>

var uuid = require('node-uuid');
var stream = require('stream');
var async = require('async');
var loader = require('./loader.js');
var cache = require('./cache.js');
var util = require('util');
var events = require('events');
var tag = 'proactive/orchestrator.js';

// Events raised by this object:
//  * 'refresh' : raised when proactive content is refreshed
function ProactiveExperienceService(clientExperienceProviders) {
  events.EventEmitter.call(this);
  this._clientExperienceProviders = clientExperienceProviders; 
  return (this);
}
util.inherits(ProactiveExperienceService, events.EventEmitter);

// Returns proactive answers and writes the base page to basePageStream
//
// params:
//  * viewDimension:  dimensions of the view port
//  * viewTheme:      current view port theme background vs accent colors)
//  * appFormCode:    [temp solution] holds the form code which was used to launch Cortana app
//  * userAgent:      [temp solution]
// errors:
//  1) 'NoConnectionError' - the current client can't reach the internet.
//  2) 'RequestTimeoutError' - request to bing.com/proactive had timed out.
ProactiveExperienceService.prototype.getAnswers = function getAnswers(params, basePageStream, cb) {
  var self = this;

  // Query cache
  console.log('%s: Looking for cached answers', tag);
  cache.getCachedAnswers(basePageStream, function processCachedAnswers(err, answers) {
    
    if (!err && answers) {
      console.log('Returning cached answers');
      return cb(null, answers);
    }
    else if (err.name !== 'CacheEmptyError') {
      // TODO This shouldn't happen. Bail out?
      console.error('%s: Error fetching from cache: ', tag, err);
    }
    
    console.log('%s: Cache is empty. Refreshing answers.', tag);
    
    // Simultaneously pipe the base page to the response stream and to a memory stream for caching
    var forkedStream = new stream.PassThrough();
    
    var basePageChunks = [];
    var cacheStream = new stream.Writable();
    cacheStream._write = function (chunk, encoding, cb) {
      basePageChunks.push(chunk);
      cb();
    }
    
    forkedStream.pipe(basePageStream);
    forkedStream.pipe(cacheStream);
    
    // Fetch new answers, validate, rank
    // Note: When Ranker is updated to consume user context (time, geo, etc), 
    //       we should apply it per impression rather than per fetch
    async.waterfall([
      async.apply(loader.fetchAllAnswers, params, forkedStream, self._clientExperienceProviders),
      validateAnswers
    //  ,rankAnswers
    ], function updateCacheAndReturnAnswers(err, answers) {
      if (err) {
        return cb(err);
      }

      // Update cache with new data. 
      // TODO Could return results immediately, but for now wait until this operation is done
      cache.setCachedAnswers(Buffer.concat(basePageChunks), answers, function returnAnswers(err) {        
        // Cache update is best effort - log and return results anyway
        if (err) {
          console.error('%s: Error during cache update: ', tag, err);
        }

        // Notify listners that we have fresh answers
        self.emit('refresh', answers);
        
        // Return the answers to our caller
        return cb(null, answers);
      });
    });
  });
}

function validateAnswers(answers, cb) {

  // TODO Refactor validation code from RankAnswers  
  if (!answers.BasePage || !Array.isArray(answers.Answers) || answers.Answers.length < 1) { 
    return cb(new Error('Answer validation failed'));
  }

  return cb(null, answers);
}

function rankAnswers(results, cb) {
  var answers = results.Answers;
  console.log('%s: Ranking %d answers', tag, answers.length);

  // Merge the client-generated and server-generated answers
  var topScore = 12.0;
  var unrankedAnswers = {};
  var unrankedCards = [];
  
  for (var i = 0; i < answers.length; i++) {
    var answer = answers[i];

    if (answer instanceof Object) {
      // Ensure that all answers have a unique Id
      if (!answer.Metadata) {
        answer.Metadata = {};
      }
      if (!answer.Metadata.Id) {
        answer.Metadata.Id = uuid.v4();
      }

      // Store answers based on their unique Id so they can be easily retrieved later
      unrankedAnswers[answer.Metadata.Id] = answer;
      //unrankedCards.push(card);
    }
  }

  return cb(null, unrankedAnswers);
}

// Export the constructor
exports.ProactiveExperienceService = ProactiveExperienceService;