/// <disable>JS2043.RemoveDebugCode</disable>
var url = require('url');
var util = require('util');
var collector = null;
var settings = require('cortana.settings');
var settingsContainer = new settings.SettingsContainer();
var tag = 'request-metrics.js';

// request ID attached to req.__reqId to correlate a request in tracing. 
// this number is chose to be unique enough for good filtering on '[xxx]'
var nextReqId = 133700000;

try {
  if (settingsContainer.readBoolValue('Proactive', 'CollectRequestMetrics')) {
    collector = require('metrics-collector');
  }
}
catch (err) {
  console.warn('CollectRequestMetrics not specified - ignoring', err);
}

//
// a middleware for measuring and logging request durations
//
function RequestMetricsMiddleware(req, res, next) {
  wrapRequest(req, res);
  return next();
}

function parseCookies(cookies, id) {
  if (typeof cookies == 'undefined') {
    return null;
  }
  var cookiesArray = cookies.split('; ');
  var cookiesMap = {};
  for (var i in cookiesArray) {
    var pair = cookiesArray[i].split('=');
    if (pair.length !== 2) {
      console.error('%s: [%d] Failed to parse request cookies: %s', tag, id, cookies);
      return null;
    }
    cookiesMap[pair[0].toLowerCase()] = pair[1];
  }
  return cookiesMap;
}

function wrapRequest(req, res) {
  // attach a unique request id.
  req.__reqId = nextReqId++;

  // check if the console is disabled and the metrics should not be collected
  if (!console.log.isEnabled && !collector) {
    return;
  }

  var _origEnd = res.end;
  var reqStartTime = Date.now();
  var cookies = parseCookies(req.headers.cookie, req.__reqId);
  var activityId = cookies ? cookies['x-activity-id'] : undefined;
  var activityStartTime = cookies ? cookies['x-activity-timestamp'] : undefined;
  var parsedUrl = url.parse(req.url);
    
  if (typeof activityStartTime != 'undefined') {
    activityStartTime = parseInt(activityStartTime);
  }
  console.log('%s: [%d] incoming Request For [%s], activity start time: [%d], start time: [%d]', tag, req.__reqId, req.url , (activityStartTime || -1), reqStartTime);
  
  // go over the headers only if the logger is enabled
  if (console.log.isEnabled) {
    var requestHeaders = '';
    Object.keys(req.headers).forEach(function (key) {
      requestHeaders += key + ': ' + req.headers[key] + '    ';
    });

    console.log('%s: [%d] request headers: %s', tag, req.__reqId, requestHeaders);
  }

  res.end = function () {
    var retVal = _origEnd.apply(res, arguments);
    
    var now = Date.now();
    var reqDuration = now - reqStartTime;
    var timeFromActivityStart = (typeof activityStartTime != 'undefined') ? (now - activityStartTime) : null;
  
    if (collector) {
      var memoryData = {'timestamp' : now,
                        'type' : collector.MEMORY,
                        'domain' : parsedUrl.pathname,
                        'id' : activityId,
                        'intValue' : process.memoryUsage().rss};

      collector.insertData(memoryData, function(err) {
        if (err) {
          console.error('%s: [%d] Failed to save memory data to metrices database', tag, req.__reqId, err);
        }
      });

      var requestData = {'timestamp' : now,
                         'type' : collector.REQUEST,
                         'domain' : parsedUrl.pathname,
                         'id' : activityId,
                         'intValue' : reqDuration,
                         'metadata' : {'timeFromActivityStart' : timeFromActivityStart} }; 

      collector.insertData(requestData, function(err) {
        if (err) {
          console.error('%s: [%d] Failed to save request data to metrices database', tag, req.__reqId, err);
        }
      });
    }
    
    // trace the request times
    console.log('%s: [%d] request duration time: [%d]', tag, req.__reqId, reqDuration);
    if (timeFromActivityStart) {
      console.log('%s: [%d] time from activity [%s] start: [%d]', tag, req.__reqId, activityId, timeFromActivityStart);
    }
        
    if(res._header) {
    	console.log('%s: [%d] response headers: %s', tag, req.__reqId, res._header.replace(/\r\n/g, '    '));
	}
    
    if (res.statusCode === 200) {
      console.log('%s: [%d] [%s] complete with status: %s', tag, req.__reqId, req.url, res.statusCode);
    }
    else if (res.statusCode < 400) {
      console.warn('%s: [%d] [%s] complete with status: %s', tag, req.__reqId, req.url, res.statusCode);
   }
    else {
      console.error('%s: [%d] [%s] failed with status: %s', tag, req.__reqId, req.url, res.statusCode);
    }
    
    return retVal;
  }
}

exports.wrapRequest = wrapRequest;
exports.RequestMetricsMiddleware = RequestMetricsMiddleware;