// Copyright(c) Microsoft Corporation. All rights reserved.
var middler = require("middler");
var app = middler();
var java = require('java');
var util = require('util');
var request = require('request');
var http = require('http');
var uuid = require('node-uuid');
var graph = require('Microsoft.Bing.Client.Graph');
var logger = require('logger')(__filename);

/**
 * adds or updates a place to user favorite places
 * @param placeInfo info for the place to be added or updated
 * @param cb callback
 * @return {*}
 */
function insertPlace(placeInfo, cb) {

    if (!placeInfo.OriginalName) {
        return cb(new Error("missing required place info: OriginalName"));
    }
    if (!placeInfo.DisplayCoordinates) {
        return cb(new Error("missing required place info: DisplayCoordinates"));
    }

    var place = null;
    var id = null;
    if (placeInfo.Id) {
        place = getPlaceById(placeInfo.Id);
        if (place) {
            // There is no way to unlink home or work from an existng place
            // we remove and add it again here
            graph.Me.Preferences.Favorites.Collections.Default.Members.Remove(place);
            place.Delete();
        }
        else {
            logger.warn("no place found with id: ", placeInfo.Id);
        }

        // get the edited place previous id in order to create a place with same id
        var idPrefix = "place.";
        id = placeInfo.Id.substring(placeInfo.Id.indexOf(idPrefix) + idPrefix.length, placeInfo.Id.length - 1);
    }

    // create a new place
    place = graph.Me.Preferences.Favorites.Collections.Default.CreatePlace(id);

    place.OriginalName = graph.CreateText(placeInfo.OriginalName);
    place.DisplayCoordinates = graph.CreateGeometryPoint(placeInfo.DisplayCoordinates.Latitude, placeInfo.DisplayCoordinates.Longitude);
    place.BingEntityId = graph.CreateText(placeInfo.BingEntityId);

    if (placeInfo.Address) {
        place.Address = graph.CreateText(placeInfo.Address);
    }
    
    if (placeInfo.Name) {
        place.Name = graph.CreateText(placeInfo.Name);
    }
    
    if (placeInfo.BusinessType) {
        place.BusinessType = graph.CreateText(placeInfo.BusinessType);
    }
    
    if (placeInfo.Type === "home") {
        graph.Me.Preferences.Favorites.SetPreferredPlace(graph.Entities.FavoritesCategoryHome, place);
    }
    else if (placeInfo.Type === "work") {
        graph.Me.Preferences.Favorites.SetPreferredPlace(graph.Entities.FavoritesCategoryWork, place);
    }

    syncGraph(function (err) {
        if (err) {
            cb(err);
        }
        cb(null, place.Id);
    });
}

/**
 * gets an array of favorite places
 * @param cb
 */
function getPlaces(cb) {
    syncGraph(function (err) {

        if (err) {
            return cb(err);
        }

        var defaultCollection = graph.Me.Preferences.Favorites.Collections.Default;

        var home = graph.Me.Preferences.Favorites.GetPreferredPlace(graph.Entities.FavoritesCategoryHome);
        var work = graph.Me.Preferences.Favorites.GetPreferredPlace(graph.Entities.FavoritesCategoryWork);

        var locations = [];
        var members = defaultCollection.Members;
        for (var i = 0; i < members.Count; i++) {
            var item = members.GetAt(i);
            if (item === undefined || item.Type !== item.Graph.Entities.Place) {
                continue;
            }

            var place = {};

            place.Id = item.Id;

            if (item.BingEntityId && item.BingEntityId.Value) {
                place.BingEntityId = item.BingEntityId.Value;
            }

            if (item.Address) {
                place.Address = (item.Address.AddressSingleLine && item.Address.AddressSingleLine.Value) || item.Address.Value;
            }

            if (item.Phone !== undefined) {
                place.Phone = item.Phone.Value;
            }

            if (item.OriginalName !== undefined) {
                place.OriginalName = item.OriginalName.Value;
            }

            if (item.Name !== undefined) {
                place.Name = item.Name.Value;
            }

            if (item.BusinessType !== undefined) {
                place.BusinessType = item.BusinessType.Value;
            }

            if (item.Geometry !== undefined) {
                place.Geometry = item.Geometry.Value;
            }

            if (item.DisplayCoordinates !== undefined) {
                place.DisplayCoordinates = { Latitude: item.DisplayCoordinates.Latitude.Value, Longitude: item.DisplayCoordinates.Longitude.Value };
            }

            if (item === home) {
                place.Type = "home";
            }
            else if (item === work) {
                place.Type = "work";
            }
            else {
                place.Type = "other";
            }

            locations.push(place);
        }

        return cb(null, locations);
    });
}

/**
 * deletes an existing place from user favorite places
 * @param id place id
 * @param cb callback
 */
function removePlace(id, cb) {
    var place = getPlaceById(id);

    if (place) {
        logger.log("removing a place with id: " + id);
        graph.Me.Preferences.Favorites.Collections.Default.Members.Remove(place);
        place.Delete();
        syncGraph(cb);
    } else {
        logger.warn("no place found with id: " + id);
        cb();
    }
}

/**
 * returns a place given its id
 * @param id place id
 * @return {*}
 */
function getPlaceById(id) {
    var members = graph.Me.Preferences.Favorites.Collections.Default.Members;
    for (var i = 0; i < members.Count; i++) {
        var item = members.GetAt(i);
        if (item && item.Id === id) {
            return item;
        }
    }
}

/**
 * syncs bing graph
 * @param cb callback
 */
function syncGraph(cb) {
    java.getAuthenticationTokens(function (err, tokens) {

        if (err) {
            logger.error('failed to get authentication tokens');
            return cb(err);
        }

        logger.log('got authentication tokens');
        var language = java.device.appUtils.getAppLanguage();
        var cookie = '_FS=mkt=' + language + '&ui=' + language;
        logger.log('calling graph.Sync with cookie', cookie);

        graph.Sync(function (statusCode) {
            if (statusCode !== 200) {
                logger.error('Failed to sync places');
                return cb(new Error("status code is " + statusCode));
            }
            // Note: this is an async call.
            logger.log("Sync completed.");
            cb();
        }, tokens.rpsToken, cookie);
    }
  )
}

// we only expose the APIs that uses the bridge if we are running on the android device,
// so that we can still run the service on the PC while developing.
if (java.isAndroid) {

    /**
     * Public create reminder endpoint by passing in a valid JSON string
     * Refer to the design spec for the JSON format
     */
    app.get('/', function (req, res) {
        logger.log('get places called');
        //var name = "Space Needle";
        //var location = {Latitude: 47.620491, Longitude: -122.349319};

        getPlaces(function (err, places) {
            if (err) {
                logger.error('failed to get places', err);
                res.writeHead(500);
                return res.end();
            }

            res.writeHead(200, { "Content-Type": "application/json; charset=utf-8" });
            var result = { locations: places };
            return res.end(JSON.stringify(result));
      });
    })

    /**
     * Create a new graph object so that cache will be clear, will trigger when user sign out
     */
    app.get('/reset', function (req, res) {
        logger.log('reset graph called');
        try {
            //delete graph object
            delete require.cache[require.resolve('Microsoft.Bing.Client.Graph')];
            //recreate a graph object
            graph = require('Microsoft.Bing.Client.Graph');

            res.writeHead(200);
            return res.end();
        } catch(ex) {
            res.writeHead(500);
            return res.end();
        }
    })

    app.post('/', function (req, res) {
        logger.log('add place called');

        insertPlace(req.body, function (err, newPlaceId) {
            if (err) {
                logger.error('failed to add place: ', err);
                res.writeHead(500);
                return res.end();
            }
            logger.log('added place successfully');
            res.writeHead(200, { "Content-Type": "application/json" });
            return res.end(newPlaceId);
        });
    })

    app.delete('/:id', function (req, res) {
        logger.log('delete a place called');

        var id = decodeURIComponent(req.params.id);
        removePlace(id, function (err) {
            if (err) {
                logger.error('failed to delete place with id', id, err);
                res.writeHead(500);
                return res.end();
            }

            logger.log('deleted place successfully');
            res.writeHead(200, { "Content-Type": "application/json" });
            return res.end();
        });
    })
}

module.exports = app;
