// Copyright(c) Microsoft Corporation. All rights reserved.

var java = require('java');
var http = require('http');
var request = require('request');
var logger = require('logger')(__filename + ' reminderSync');
var util = require('util');
var dssutil = require('dssutil');
var async = require('async');
var agent = require('./agentConfig.js');

var REMINDER_AGENT_ID = 'Reminder_2.0.0.0.0';
var QUERY_FORMAT = 'query=%s&entity=&url=&entityType=&spatialFilter=';
var GEOFENCE_GROUP_MODE = '1';
var GEOFENCE_COUNT = '10';
var GEOFENCE_ENDDATE = '2114-05-17 13:40:13-07:00';
var GEOFENCE_PARENT_ID = '{BF38793E-EE2E-4C4E-98CB-B580418FD429}'; // fake id as it is not used but needs by cloud
var AIS_APP_ID = 'e2b98658-56ac-45f0-87f0-4acaf12f3a51';

function BusinessLocationQuery(options) {
  if (!(this instanceof BusinessLocationQuery)) {
    return new BusinessLocationQuery(options);
  }
  
  this.appid = AIS_APP_ID;
  this.agentid = REMINDER_AGENT_ID;
  this.agentdata = {
    Latitude: options.latitude,
    Longitude: options.longitude,
    ReminderId: options.reminderId,
    LocationHint: util.format(QUERY_FORMAT, options.name),
    GeofenceGroupMode: GEOFENCE_GROUP_MODE,
    Count: GEOFENCE_COUNT,
    EndDateTime: GEOFENCE_ENDDATE,
    FirstCall: options.firstCall,
    ParentId: GEOFENCE_PARENT_ID
  };

  return [this];
}

function GeofenceResult(serverGeofence) {
  if (!(this instanceof GeofenceResult)) {
    return new GeofenceResult(serverGeofence);
  }

  this.latitude = serverGeofence.Latitude;
  this.longitude = serverGeofence.Longitude;
  this.radius = serverGeofence.Radius;
}

/**
 * Create Instance call to retrieve the store list base on the user current location, and then combine the cloud
 * JSON response with the request response to create the reminder.
 *
 * @param options options.reminder contains the reminders. options.userAgent contains the user agent
 * @param cb  call back function
 * @returns {*} return the agent cloud json response for create instance that combines with the request json
 */
function businessToGeofence(options, cb) {
  var device = require('device');
  
  cb = cb || Function();
  
  if (!options ||
      !options.name ||
      !options.reminderId) {
    return dssutil.cbError(cb, logger, 'invalid arg: %j', options);
  }
  
  return async.parallel(
    [
      function getHeaders(parallelCb) {
        device.getRemindersHeaders(parallelCb);
      },

      function getCurrentLocation(parallelCb) {
        var locationCb = java.createJavaProxy('com.microsoft.bing.dss.platform.location.location.ILocationListener', {
          onLocation : function (location) {
            parallelCb(null, { latitude: location.getLatitude(), longitude : location.getLongitude() });
          },
          onRequestError : function (statusCode) {
            parallelCb(new Error('failed to get current location: ' + statusCode));
          }
        });
        
        return java.device.locationApi.requestCurrentLocation(locationCb);
      }
    ],

    function asyncComplete(err, results) {
      if (err) {
        return dssutil.cbError(cb, logger, err);
      }
      options.headers = results[0];
      options.latitude = results[1].latitude;
      options.longitude = results[1].longitude;
      return _queryForBusiness(options, cb);
    });
}

function _queryForBusiness(options, cb) {
  var payload = new BusinessLocationQuery(options);
  logger.log('querying for business location with payload: %j', payload);
  return request.post(agent.getBusinessGeofenceServiceUrl(), { json: payload, headers: options.headers }, function (err, response, body) {
    if (err) {
      return dssutil.cbError(cb, logger, err);
    }
    
    if (response.statusCode != 200 || !body) {
      return dssutil.cbError(cb, logger, '{ "status": %s, "entity": %j }', response.statusCode, body);
    }
    
    if (body.success !== 'true') {
      return dssutil.cbError(cb, logger, 'createInstance reported error, entity: %j', body);
    }
    
    logger.log('CreateInstance succeeded: %j ', response.headers);
    logger.verbose('geofence payload: %j', body);
    
    var geofenceResults = {
      outerGeofences : [],
      innerGeofences : []
    };
    
    var rules = body.rules;
    rules.forEach(function (rule) {
      if (!rule.ruleCode) {
        logger.error('ruleCode not found.: %j', rule);
        return;
      }
      
      var ruleJson;
      // for some reason ruleCode is a string of JSON...
      try {
        ruleJson = JSON.parse(rule.ruleCode);
      }
      catch (e) {
        logger.error('ruleCode parsing failed ', rule.ruleCode, e);
        return;
      }

      var outerGeofence = ruleJson.OuterGeofence;
      var innerGeofences = ruleJson.InnerGeofences;
      
      if (!outerGeofence || !innerGeofences) {
        logger.error('outerGeofence/innerGeofences not found.: %j', rule);
        return;
      }

      geofenceResults.outerGeofences.push(new GeofenceResult(outerGeofence));
      innerGeofences.forEach(function (innerGeofence) {
        geofenceResults.innerGeofences.push(new GeofenceResult(innerGeofence));
      });
      
      return;
    });

    cb(null, geofenceResults);
  });
}

exports.businessToGeofence = businessToGeofence;