// Copyright(c) Microsoft Corporation. All rights reserved.

var platform_reminder = require('./platform-reminder.js');

var logger = require('logger')(__filename + ' reminderSync');
var dssutil = require('dssutil');

var types = require('./types.js');
var handlers = require('./handlers.js');
var trigger = require('./trigger.js');

function _createReminder(reminderManifest, synced, cb) {
  var reminderDescriptor = platform_reminder.createReminderDescriptor();
  
  reminderDescriptor.setSyncStatus(synced)
    .setStatus(reminderManifest.reminderStatus);

  if (reminderManifest.id) {
    // The create is coming from the server - update the serverId.
    reminderDescriptor.setServerId(reminderManifest.id);
  }
  
  var handler = handlers.typeToHandler(reminderManifest.reminderType);
  return handler.createReminder({ manifest: reminderManifest }, reminderDescriptor, cb);
}

function _updateReminder(reminderManifest, reminderDescriptor, synced, cb) {
  reminderDescriptor.setSyncStatus(synced)
    .setStatus(reminderManifest.reminderStatus);
  
  var extraJson;

  // Get the existing state we have stored with the reminder.
  try {
    extraJson = JSON.parse(reminderDescriptor.getExtraData());
  }
  catch (e) {
    return dssutil.cbError(cb, logger, e);
  }

  var existingManifest = extraJson.manifest;
  
  if (existingManifest === undefined) {
    // TODO: recover from this?
    return dssutil.cbError(cb, logger, 'unexpected missing data in reminder Db');
  }

  var existingLastUpdate = new Date(existingManifest.lastUpdatedAt).getTime();
  var newLastUpdate = new Date(reminderManifest.lastUpdatedAt).getTime();
  
  if (newLastUpdate === undefined) {
    return dssutil.cbError(cb, logger, 'invalid date newLastUpdate: ' + reminderManifest.lastUpdatedAt);
  }

  if (existingLastUpdate !== undefined && (existingLastUpdate > newLastUpdate)) {
    // We have a local version newer than the incoming update.
    return cb();
  }
  
  var handler = handlers.typeToHandler(reminderManifest.reminderType);
  
  // Unregister the reminder handlers so they can be updated.
  handler.unregisterReminder(extraJson, reminderManifest, function () {
    if (reminderManifest.reminderStatus === types.ReminderStatus.Cancelled &&
        (synced || !reminderManifest.id)) {
      //The reminder is cancelled and (synced or never synced originally), just remove it.
      return platform_reminder.deleteByLocalId(reminderDescriptor.getLocalId(), cb);
    }

    extraJson.manifest = reminderManifest;

    // Do a full recreate.
    return handler.createReminder(extraJson, reminderDescriptor, function (err, id) {
      if (err) {
        return dssutil.cbError(cb, logger, 'error updating reminder');
      }
      
      // check whether we have a new snoozed state and set the snooze alarm.
      if (reminderManifest.reminderStatus === types.ReminderStatus.Snoozed &&
          existingManifest.reminderStatus !== types.ReminderStatus.Snoozed) {
        trigger.snoozeReminder(id);
      }

      return cb(null, id);
    });
  });
}

function createOrUpdateReminder(reminderManifest, synced, localId, cb) {
  cb = cb || Function();
  
  if (!reminderManifest) {
    return dssutil.cbError(cb, logger, 'invalid parameter reminderManifest');
  }

  // Determine whether we have this reminder already, preferring the serverId.
  var queryFn;
  if (reminderManifest.id && reminderManifest.id !== '') {
    queryFn = platform_reminder.queryByServerId.bind(null, reminderManifest.id);
  }
  else if (localId !== undefined) {
    queryFn = platform_reminder.queryByLocalId.bind(null, localId);
  }
  
  return queryFn(function (err, existing) {
    // Do the create or update.
    if (!err) {
      return _updateReminder(reminderManifest, existing, synced, cb);
    }
    else {
      return _createReminder(reminderManifest, synced, cb);
    }
  });
}

function deleteReminder(localId, serverId, synced, cb) {
  cb = cb || Function();
  var queryFn;
  // Find the reminder in the db.
  if (serverId) {
    queryFn = platform_reminder.queryByServerId.bind(null, serverId);
  }
  else if (localId) {
    queryFn = platform_reminder.queryByLocalId.bind(null, localId);
  }
  
  return queryFn(function (err, reminderDescriptor) {
    if (err) {
      return dssutil.cbError(cb, logger, 'reminder not found, local:%d, server:%d ', localId, serverId);
    }
    
    // Update the common fields, marking the reminder as CANCELLED - update will handle this as deleted.
    var reminderManifest = JSON.parse(reminderDescriptor.getExtraData()).manifest;
    reminderManifest.lastUpdatedAt = new Date().toISOString();
    reminderManifest.lastViewedAt = new Date().toISOString();
    reminderManifest.reminderStatus = types.ReminderStatus.Cancelled;
    
    return _updateReminder(reminderManifest, reminderDescriptor, synced, cb);
  });
}

function readReminder(queryId, cb) {
  cb = cb || Function();
  logger.verbose('readReminder');
  
  var reminders;
  
  // Pull the reminder(s) from the db.
  if (!queryId) {
    return platform_reminder.queryAll(function (err, reminders) {
      if (err) {
        return dssutil.cbError(cb, logger, err);
      }
      
      return processReminders(reminders, cb);
    });
  }

  return platform_reminder.queryByLocalId(queryId, function (err, reminder) {
    if (!reminder) {
      return dssutil.cbError(cb, logger, 'reminder not found, queryId:%d ', queryId);
    }

    processReminders([reminder], cb);
  });
}

function processReminders(reminders, cb) {
  var results = [];
  var extraData;
  var localId;
  var manifest;
  for (var i = 0; i < reminders.length; i++) {
    localId = reminders[i].getLocalId();
    try {
      extraData = JSON.parse(reminders[i].getExtraData());
    }
    catch (e) {
      logger.error('unexpected data in reminderDb: ', localId);
      continue;
    }
    
    if (extraData.manifest === undefined) {
      logger.error('missing manifest data in reminderDb: ', localId);
      continue;
    }
    results.push({
      localId: localId,
      manifest: extraData.manifest
    });
  }

  cb(null, results);
}

function clearReminderDb(req, res) {
    return platform_reminder.clearLocalReminderDb(function (err) {
        if(err) {
            res.writeHead(500);
            return res.end();
        } else {
            res.writeHead(200);
            return res.end();
        }
    });
}

exports.createOrUpdateReminder = createOrUpdateReminder;
exports.readReminder = readReminder;
exports.deleteReminder = deleteReminder;
exports.clearReminderDb = clearReminderDb;