// Copyright(c) Microsoft Corporation. All rights reserved.

var platform_reminder = require('./platform-reminder.js');

var logger = require('logger')(__filename + ' reminderSync');
var dssutil = require('dssutil');

var types = require('./types.js');
var v1Types = require('./v1types.js');
var crudV2 = require('./crud.js');
var handlers = require('./handlers.js');

function createReminderFromV1(v1Obj, cb) {
  cb = cb || Function();
  logger.verbose('create payload: %j', v1Obj);
  
  if (!v1Obj) {
    return dssutil.cbError(cb, logger, 'invalid parameter v1Obj');
  }
  
  // create the new v2 reminder manifest for storing locally.
  var reminderManifest = new types.ReminderManifest();
  reminderManifest.createdAt = new Date().toISOString();
  reminderManifest.lastUpdatedAt = reminderManifest.createdAt;
  reminderManifest.lastViewedAt = reminderManifest.lastUpdatedAt;
  reminderManifest.reminderStatus = types.ReminderStatus.Active;
  
  return _createOrUpdateFromV1(v1Obj, reminderManifest, cb);
}

function updateReminderFromV1(v1Obj, cb) {
  cb = cb || Function();
  logger.verbose('update payload: %j', v1Obj);
  
  if (!v1Obj) {
    return dssutil.cbError(cb, logger, 'invalid parameter v1Obj');
  }
  
  // Based on the type get the local reminderId.
  // TODO: design a local format where id is in the top level object not on each type.
  var reminderId;
  if (v1Obj.timeReminder) {
    reminderId = v1Obj.timeReminder[0].reminderId;
  }
  else if (v1Obj.businessLocationReminder) {
    reminderId = v1Obj.businessLocationReminder[0].reminderId;
  }
  else if (v1Obj.locationReminder) {
    reminderId = v1Obj.locationReminder[0].reminderId;
  }

  if (!reminderId) {
    return dssutil.cbError(cb, logger, 'invalid reminder: %j', v1Obj);
  }
  
  // For update we need pass the cloud the existing serverId, pull it from the db.
  return platform_reminder.queryByLocalId(reminderId, function (err, reminderDescriptor) {
    if (err) {
      return dssutil.cbError(cb, logger, 'reminder not found: ', reminderId);
    }
    
    var reminderManifest = JSON.parse(reminderDescriptor.getExtraData()).manifest;
    reminderManifest.lastUpdatedAt = types.getDateIsoString();
    reminderManifest.lastViewedAt = reminderManifest.lastUpdatedAt;
    
    return _createOrUpdateFromV1(v1Obj, reminderManifest, cb);
  });
}

function deleteReminderFromV1(v1Obj, cb) {
  cb = cb || Function();
  logger.verbose('delete payload: %j', v1Obj);
  
  if (!v1Obj) {
    return dssutil.cbError(cb, logger, 'invalid parameter v1Obj');
  }

  return crudV2.deleteReminder(v1Obj.reminderId, undefined, false, cb);
}

function _createOrUpdateFromV1(v1Obj, reminderManifest, cb) {
  var localId;
  var handler;
  var detailObj;

  // At this point the remindermanifest is partially filled, fill in the remaining fields.
  if (v1Obj.timeReminder) {
    handler = handlers.typeToHandler(types.ReminderType.Time);
    detailObj = v1Obj.timeReminder[0];
  }
  else if (v1Obj.businessLocationReminder) {
    handler = handlers.typeToHandler(types.ReminderType.Location);
    detailObj = v1Obj.businessLocationReminder[0];
    detailObj.type = 'business';
  }
  else if (v1Obj.locationReminder) {
    handler = handlers.typeToHandler(types.ReminderType.Location);
    detailObj = v1Obj.locationReminder[0];
    detailObj.type = 'location';
  }
  else {
    return dssutil.cbError(cb, logger, 'invalid reminder type: ', JSON.stringify(v1Obj));
  }

  if (!handler.parseFromV1(reminderManifest, detailObj)) {
    return dssutil.cbError(cb, logger, 'invalid reminder object: %j ', v1Obj);
  }

  localId = detailObj.reminderId;
  
  return crudV2.createOrUpdateReminder(reminderManifest, false, localId, cb);
}

function readReminderAsV1(queryId, cb) {
  cb = cb || Function();
  return crudV2.readReminder(queryId, function (err, reminders) {
    if (err) {
      return dssutil.cbError(cb, logger, err);
    }
    
    // Convert the reminder to the V1 format.
    var v1Array = [];
    for (var i = 0; i < reminders.length; i++) {
      var manifest = reminders[i].manifest;
      var localId = reminders[i].localId;
      
      if (localId === undefined ||
          manifest.title === undefined ||
          manifest.description === undefined ||
          v1Types.ReminderStatusV2toV1Map[manifest.reminderStatus] === undefined) {
        logger.error('cannot convert from v2 to v1: %j', manifest);
        continue;
      }

      var reminder = {
        reminderId : localId,
        title : manifest.title,
        detail : manifest.description,
        status : v1Types.ReminderStatusV2toV1Map[manifest.reminderStatus]
      };
      
      // Parse the specific reminder type and push it onto results.
      var handler = handlers.typeToHandler(manifest.reminderType);
      if (!handler.parseToV1(reminder, v1Array, manifest)) {
        logger.error('cannot convert from v2 to v1: %j', manifest);
        continue;
      }
    }

    cb(null, queryId ? v1Array[0] : v1Array);
  });
}

function clearReminderDb(req, res) {
    return crudV2.clearReminderDb(req, res);
}

exports.createReminderFromV1 = createReminderFromV1;
exports.updateReminderFromV1 = updateReminderFromV1;
exports.deleteReminderFromV1 = deleteReminderFromV1;
exports.readReminderAsV1 = readReminderAsV1;
exports.clearLocalReminderDb = clearReminderDb;