// Copyright(c) Microsoft Corporation. All rights reserved.

var middler = require('middler');
var app = middler();
var java = require('java');
var util = require('util');
var request = require('request');
var http = require('http');
var logger = require('logger')(__filename + ' reminderSync');
var dssutil = require('dssutil');

var crudV1 = require('./crudv1.js');
var sync = require('./sync.js');
var business = require('./business.js');

// we only expose the APIs that uses the bridge if we are running on the android device,
// so that we can still run the service on the PC while developing.
if (java.isAndroid) {
    
  function onCreateUpdateDeleteComplete(req, res, err, result) {
    if (err) {
      dssutil.logError(err);
      return completeWithError(res, err, req.url);
    }

    // initiate sync with the cloud, sync will post an event to the device when complete.
    sync.syncReminders(true);
    return completeWithReminderId(res, result, req.url);    
  }

  function completeWithError(res, error, url) {
    try {
        // parse error content as JSON object
        error = JSON.parse(error.message);

        res.writeHead(error.status, { 'Content-Type': 'application/json' });

        logger.info(url + ' complete with result: %j', error);
        return res.end(JSON.stringify(error));
    }
    catch (exception) {
        // don't let the app crash due to unable to parse "error" string/data.
        logger.error('error while parsing response data');
        var customError = {
            "status":400,
            "entity":{
              "success":"false",
              "error":{
                 "code":"ResponseParseError",
                 "message":"Failed to parse server response."
              }
            }
        }
        return res.end(JSON.stringify(customError));
    }
  }

  function completeWithReminderId(res, reminderId, url) {
    var json = { 'success': reminderId ? true : false, 'reminderId': reminderId };
    res.writeHead(200, { 'Content-Type': 'application/json' });
    
    logger.info(url + ' complete with result: %j', json);
    return res.end(JSON.stringify(json));
  }
  
  function onGetComplete(req, res, err, result) {
    res.writeHead(200, { 'Content-Type': 'application/json' });
    logger.info(req.url + ' complete with result 200 OK - %j', result);
    return res.end(JSON.stringify(result));
  }

  /**
    * Public create reminder endpoint by passing in a valid JSON string
    * Refer to the design spec for the JSON format
    */
  app.post('/', function (req, res) {
    var jsonObj = req.body;

    jsonObj.reminderId = undefined;
    return crudV1.createReminderFromV1(jsonObj, onCreateUpdateDeleteComplete.bind(undefined, req, res));
  });
  
  /**
    * Public update endpoint to update the reminder.
    * Note that regular update is for field title/detail
    * In case for business reminder if the user change the location as update, we have to delete and recreate a new reminder
    */
  app.post('/update', function (req, res) {
    var jsonObj = req.body;
    return crudV1.updateReminderFromV1(jsonObj, onCreateUpdateDeleteComplete.bind(undefined, req, res));
  });

  /**
    * Public delete endpoint to delete the reminder by reminder id
    */
  app.post('/delete/:id', function (req, res) {
    // call DLIB deleteReminder() by passing the JSON string
    var id = decodeURI(req.params.id);
    return crudV1.deleteReminderFromV1({ reminderId: id }, onCreateUpdateDeleteComplete.bind(undefined, req, res));
  });
  
  /**
    * Public get endpoint to retrieve the reminder by JSON request string, mostly with reminder id
    */
  app.post('/get/:id', function (req, res) {
    // initiate sync with the cloud, sync will post an event to the device when complete.
    sync.syncReminders(false);
    
    // read from localDb.
    return crudV1.readReminderAsV1(decodeURI(req.params.id), onGetComplete.bind(undefined, req, res));
  });
  
  /**
    * Public get endpoint to retrieve all reminders
    */
  app.post('/get', function (req, res) {
    // initiate sync with the cloud, sync will post an event to the device when complete.
    sync.syncReminders(false);

    // read from localDb.
    return crudV1.readReminderAsV1(undefined, onGetComplete.bind(undefined, req, res));
  });

  /**
   * Public clear local database for user sign out
   */
  app.get('/clear', function(req, res) {

      logger.log('clear local reminders');

      //clear local database
      return crudV1.clearLocalReminderDb(req, res);
  });
}

module.exports = app;