﻿// Copyright(c) Microsoft Corporation. All rights reserved.

var platform_location = require('./platform-location.js');
var platform_reminder = require('./platform-reminder.js');

var types = require('./types.js');
var v1Types = require('./v1types.js');

var business = require('./business.js');
var trigger = require('./trigger.js');
var handlers = require('./handlers.js');

var logger = require('logger')(__filename + ' reminderSync');
var dssutil = require('dssutil');
var async = require('async');

var NEVER_EXPIRE = -1;

platform_location.onEnteringGeofence(function (geofenceDescriptor) {
  return handleGeofenceTransition(geofenceDescriptor.getTag());
});

platform_location.onExittingGeofence(function (geofenceDescriptor) {
  var tag = geofenceDescriptor.getTag();
  var match = tag.match(/reminder\.outerGeofence\.([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})$/);
  if (match) {

    // update the business goefence.
    var reminderId = match[1];
    return platform_reminder.queryByLocalId(reminderId, function (err, reminderDescriptor) {
      if (err) {
        logger.error('error finding outter geofence: ', err);
        return;
      }
      
      var extraJson;
      
      try {
        extraJson = JSON.parse(reminderDescriptor.getExtraData());
      }
      catch (e) {
        logger.error('reminder error: ', e);
        return;
      }
      
      var locationHandler = handlers.typeToHandler(types.ReminderType.Location);

      return async.series([
        locationHandler.unregisterReminder.bind(locationHandler, extraJson, undefined),
        locationHandler.createReminder.bind(locationHandler, extraJson, reminderDescriptor)
      ],
      function (err) {
        if (err) {
          // TODO: we need to try to recreate at a later time.
          logger.error('error unregistering reminder: %s: %j', err, extraJson);
          return;
        }
      });
    });
  }
  
  // Otherwise handle the transition for a location or inner geofence.
  return handleGeofenceTransition(tag);
});

function handleGeofenceTransition(tag) {
  var match = tag.match(/reminder\.([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})$/);
  if (!match) {
    return undefined;
  }

  var reminderId = match[1];
  trigger.triggerReminder(reminderId);
}

function LocationHandler() {
  if (!(this instanceof LocationHandler)) {
    return new LocationHandler();
  }
  
  this.parseFromV1 = function (reminderManifest, v1Obj, cb) {
    var location = new types.ReminderLocation();
    
    if (!v1Types.validate(v1Obj) ||
        v1Obj.placeName === undefined ||
        v1Obj.geofenceKind === undefined) {
      return false;
    }
    
    // Time indicates the start of the reminder.  Currently we don't support end dates.
    location.name = v1Obj.placeName;
    location.geofenceType = v1Types.GeofenceKindV1ToV2[v1Obj.geofenceKind];
    if (v1Obj.type === 'location') {
      if (v1Obj.latitude === undefined ||
          v1Obj.longitude === undefined) {
        return false;
      }
      
      //TODO: address.addressSingleLine = v1Obj.???
      location.geo = new types.ReminderGeo();
      location.geo.latitude = v1Obj.latitude;
      location.geo.longitude = v1Obj.longitude;
    }
    
    location.address = new types.ReminderAddress();
    
    // Copy to the V2 manifest.
    // because CoA supports only Location reminders we need to sync all
    // of the location reminders as location reminders not TimeLocation Reminders.
    reminderManifest.reminderType = types.ReminderType.Location;
    reminderManifest.title = v1Obj.title;
    reminderManifest.description = v1Obj.detail;
    reminderManifest.location = location;
    reminderManifest.reminderStatus = v1Types.ReminderStatusV1toV2Map[v1Obj.status];
    return true;
  };
  
  this.parseToV1 = function parseToV1(v1Reminder, results, reminderManifest) {
    // Convert from the V2 format to the V1 format.  v1Reminder is already set with the common fields.
    var location = reminderManifest.location;
    if (location === undefined) {
      return false;
    }
    v1Reminder.placeName = location.name;
    v1Reminder.geofenceKind = v1Types.GeofenceKindV2toV1Map[location.geofenceType];
    
    if (v1Reminder.placeName === undefined ||
        v1Reminder.geofenceKind === undefined) {
      return false;
    }
    
    if (!location.address) {
      return false;
    }
    
    if (location.geo) {
      v1Reminder.latitude = location.geo.latitude;
      v1Reminder.longitude = location.geo.longitude;
      if (v1Reminder.latitude === undefined || v1Reminder.longitude === undefined) {
        return false;
      }
      
      results.push({ locationReminder : [v1Reminder] });
    }
    else {
      // TODO: for some reason the v1 businessReminder expects a location here...
      //   Not sure yet that this will cause a problem or not.
      v1Reminder.latitude = 45.0;
      v1Reminder.longitude = 90.0;
      results.push({ businessLocationReminder : [v1Reminder] });
    }
    
    return true;
  };
  
  this.createReminder = function (extraJson, reminderDescriptor, cb) {
    // Set the timer on the device.
    this._setGeofenceFromManifest(extraJson, reminderDescriptor.getLocalId(), function (err, geofenceIds) {
      if (err) {
        return dssutil.cbError(cb, logger, err);
      }
      
      extraJson.geofenceIds = geofenceIds;
      logger.verbose('createData:%j', extraJson);
      
      // Use the extraData field to store the relation with the timerId and the full manifest which
      // is opaque data to the db.
      reminderDescriptor.setType(types.ReminderType.Location)
        .setExtraData(JSON.stringify(extraJson));
      
      return platform_reminder.createOrUpdateReminder(reminderDescriptor, function (err) {
        if (err) {
          return dssutil.cbError(cb, logger, 'could set reminder');
        }

        return cb(null, reminderDescriptor.getLocalId());
      });
    });
  };
  
  // unregisters the triggers if update is needed.  Completes with true if needed, false otherwise.
  this.unregisterReminder = function (existingExtraJson, newManifest, cb) {

    if (!existingExtraJson) {
      logger.verbose("existingExtraJson is undefined");
      return cb("existingExtraJson is undefined");
    }

    if (!existingExtraJson.geofenceIds) {
      logger.verbose("There are no geofence Ids to unregister");
      return cb();
    }

    if (newManifest && newManifest.reminderStatus === types.ReminderStatus.Active &&
        existingExtraJson.manifest) {
      if (types.areEqualLocations(newManifest.location, existingExtraJson.manifest.location)) {
        // The new location matches the existing location ... don't remove them.
        logger.verbose('skipping geofence delete because location matches existing %j -- %j', newManifest.location, existingExtraJson.manifest.location);
        return cb();
      }
    }

    return async.each(existingExtraJson.geofenceIds, function (geofenceId, eachCb) {
      // Remove each geofence from the Db.
      return platform_location.unregisterGeofence(geofenceId, function () { eachCb() });
    },
    function (err) {
      existingExtraJson.geofenceIds = undefined;
      return cb();
    });
  };
  
  this._setGeofenceFromManifest = function (extraJson, localId, cb) {
    var self = this;

    if (extraJson.manifest.reminderStatus !== types.ReminderStatus.Active) {
      // reminder is not active, don't set the geofence.
      return cb(null, extraJson.geofenceIds);
    }
    
    if (extraJson.geofenceIds !== undefined) {
      // already have the geofences set.
      return cb(null, extraJson.geofenceIds);
    }
    
    var location = extraJson.manifest.location;
    
    if (location.geo) {

      var geofence = location.geo;
      geofence.geofenceType = location.geofenceType;
      geofence.reminderId = localId;
      geofence.radius = location.radius;
      // basic location with geofence, just set it.
      return self._setDeviceGeofence(geofence, function (err, geofenceId) {
        if (err) {
          return dssutil.cbError(cb, logger, err);
        }

        return cb(null, [geofenceId]);
      });
    }
    
    var options = {
      name : location.name,
      reminderId : localId,
      firstTime : true //TODO: what is this for?
    };
    
    return business.businessToGeofence(options, function (err, businessGeofences) {
      if (err) {
        return dssutil.cbError(cb, logger, err);
      }

      var geofenceIds = [];
      
      return async.parallel([
        // Set all the inner geofences.
        async.each.bind(self, businessGeofences.innerGeofences, function (innerGeofence, eachCb) {
          innerGeofence.geofenceType = location.geofenceType;
          innerGeofence.reminderId = localId;
          return self._setDeviceGeofence(innerGeofence, function (err, innerId) {
            if (err) {
              return dssutil.cbError(eachCb, logger, err);
            }
            
            innerGeofence.added = true;
            geofenceIds.push(innerId);
            eachCb();
          });
        }),

        // Set all the outer geofences.
        async.each.bind(self, businessGeofences.outerGeofences, function (outerGeofence, eachCb) {
          outerGeofence.geofenceType = types.ReminderGeoFenceType.Depart;
          outerGeofence.outer = true;
          outerGeofence.reminderId = localId;
          return self._setDeviceGeofence(outerGeofence, function (err, outerId) {
            if (err) {
              return dssutil.cbError(eachCb, logger, err);
            }
            
            outerGeofence.added = true;
            geofenceIds.push(outerId);
            eachCb();
          });
        })],
        
        function (err) {
        if (err) {
          //TODO: roll back created geofences.
          return dssutil.cbError(cb, logger, err);
        }
        
        cb(null, geofenceIds);
      });
    });
  };
  
  this._setDeviceGeofence = function (geofence, cb) {
    var tag = 'reminder.';
    if (geofence.outer) {
      tag += 'outerGeofence.';
    }
    tag += geofence.reminderId;
    
    var transition = platform_location.constants.GEOFENCE_TRANSITION_ENTER;
    
    // TODO: for nearby maybe we should just increase the radius a little?
    
    if (geofence.geofenceType == types.ReminderGeoFenceType.Depart) {
      transition = platform_location.constants.GEOFENCE_TRANSITION_EXIT;
    }
    
    var geofenceDescriptor = platform_location.createGeofenceDescriptor(tag,
                                                                        '',
                                                                        geofence.latitude,
                                                                        geofence.longitude,
                                                                        geofence.radius,
                                                                        NEVER_EXPIRE,
                                                                        transition,
                                                                        0);

    return platform_location.registerGeofence(geofenceDescriptor, function (err) {
      if (err) {
        return dssutil.cbError(cb, logger, 'could not register geofence: ', err);
      }

      return cb(null, geofenceDescriptor.getId());
    });
  };
}

exports.LocationHandler = LocationHandler;