// Copyright(c) Microsoft Corporation. All rights reserved.

var platform_reminder = require('./platform-reminder.js');
var java = require('java');
var request = require('request');
var async = require('async');
var logger = require('logger')(__filename + ' reminderSync');
var dssutil = require('dssutil');

var agentConfig = require('./agentConfig.js');
var device = require('device');
var crud = require('./crud.js');
var types = require('./types.js');

// reminder operations
var REMINDER_CREATE = '/reminder/createmultiple';
var REMINDER_DELETE = '/reminder/delete';
var REMINDER_GET = '/reminder/';   //http get - get all , http post - get all details
var REMINDER_CREATEINSTANCE = '/CreateInstance';
var REMINDER_UPLOAD = '/reminder/uploaddata';

var REMINDER_ETAG_KEY = 'reminderFullSyncEtag_3';

var HALSEY_ETAG_HEADER = 'X-Halsey-ETag';

// Contract with the server is to wait 1 min. before requesting a pushed reminder update.
var PUSH_TO_PULL_DELAY = 60 * 1000;

// Wait at least 1 minute between syncs.
var MIN_SYNC_INTERVAL = 60 * 1000;

var deviceId;
var config;

var _pushHandlers = {
  'fetch' : _pushSyncById,
  'delete' : _pushDeleteById
};

// Register for node events which will include push notifications for
// ActionTypeQueueActionUris and event name pushActionUri.
var javaListener = java.eventsManager;
javaListener.start(function (event) {
  logger.info('received push event: ' + event.name);
      
  if (event.name !== 'pushActionUri' || !event.data || !event.data.actionUris) {
    return;
  }
  
  var actionUris;

  try {
    actionUris = JSON.parse(event.data.actionUris);
  }
  catch (e) {
    return logger.error('error parsing actionUris: ' + event.data.actionUris);
  }

  if (!actionUris || !actionUris instanceof Array) {
    return logger.error('invalid actionUris: %j', actionUris);
  }
  
  // Log the device ID so we don't sync pushes originating from our own device.
  deviceId = deviceId || java.device.utils.getDeviceId();
  
  return actionUris.forEach(function (actionUri) {
    if (!actionUri || !actionUri.match) {
      return logger.error('invalid actionUri: ' + actionUri);
    }
    
    logger.info('processing push notification: ' + actionUri);
  
    // find the metadata for the push. action://reminder/[fetch|delete]?id=[id]&deviceid=[deviceid]
    // TODO: Is there a URL parser available for this?
    var match = actionUri.match(/^action:\/\/reminder\/(.+)[\\?]{1}id=([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})&deviceid=([a-fA-F0-9]+)/);
    if (!match) {
      return logger.debug('actionUri not in correct format for reminders ' + actionUri);
    }
    
    if (match[3] == deviceId) {
      // The notification is for our own device.
      logger.info('ignoring push from own device ' + actionUri);
      return;
    }
    
    logger.info('performing [' + match[1] + '] on [' + match[2] + ']');
    var handler = _pushHandlers[match[1]] || function () { };
    
    // Queue the sync to the queue.
    handler(match[2]);
    return;
  });
});


// Create a queue to handle serializing sync operations one at a time.
var syncQueue = async.queue(function (task, callback) {
  logger.info('queueing sync task.');
  task(callback);
}, 1);

// Only have one outstanding full sync at a time. All others requesting full sync are put on the waiters list.
syncQueue._fullSyncInProgess = false;
syncQueue._fullSyncWaiters = [];
syncQueue._lastSyncTime;
syncQueue._pendingTimer;

syncQueue.empty = function () {
  //TODO: Notify the UX that there's an update.
  //platform_reminder.notifyChange();
};

function _pushSyncById(id) {
  setTimeout(function() {
    return syncQueue.push(function (qCb) {
      logger.info('running sync of id: ' + id);
      device.getRemindersHeaders(function (err, headers) {
        if (err) {
          return dssutil.cbError(qCb, logger, err);
        }
        
        // Request the ID from the cloud and update the local reminder.
        var payload = [id];
        return _createOrUpdateFromCloud(payload, headers, function (err) {
          if (err) {
            return dssutil.cbError(qCb, logger, err);
          }

          // Notify listeners on the device that an update occurred.
          platform_reminder.notifyReminderUpdate();
          return qCb();
        });
      });
    });
    }, PUSH_TO_PULL_DELAY);
}

function _pushDeleteById(id) {
  return syncQueue.push(function (qCb) {
    logger.log('running delete of id: ' + id);
    
    // Delete the reminder from the device.
    crud.deleteReminder(undefined, id, true, qCb);
    // Notify listeners on the device that an update occurred.
    platform_reminder.notifyReminderUpdate();
  });
}

// TODO: Add an automatic periodic sync.
// Trigger a sync of all reminders.  
// If a sync is in progress this request will be fulfilled by the outstanding sync.
// If a sync has recently completed don't start a new one.  If this sync request is
//  due to a change on the client delay trigger another sync.
// changed - true if syncing due to change. false if syncing due to fetch as a hint.
function syncReminders(changed, cb) {
  cb = cb || Function();
  
  config = config || java.getConfiguration();

  // Only allow one outstanding full sync at time.
  if (syncQueue._fullSyncInProgess) {
    logger.info('full sync waiting on result of previous sync.');
    return syncQueue._fullSyncWaiters.push(cb);
  }
  
  if (syncQueue._lastSyncTime !== undefined && ((Date.now() - syncQueue._lastSyncTime) < MIN_SYNC_INTERVAL)) {
    logger.info('skipping full sync because of recent previous sync');
    if (changed && syncQueue._pendingTimer === undefined) {
      // Schedule a sync to make sure new updates are eventually picked up.
      syncQueue._pendingTimer = setTimeout(syncReminders, MIN_SYNC_INTERVAL);
    }

    return cb();
  }

  // Mark the full sync and push this callback to execute on completion.
  syncQueue._fullSyncInProgess = true;
  syncQueue._fullSyncWaiters.push(cb);
  syncQueue._pendingTimer = undefined;

  // Remove any pushes in the queue already, we will capture them all in the sync.
  syncQueue.kill();

  // Start the full sync.
  return syncQueue.push(function (qCb) {
    logger.log('running full sync');
    
    //TODO: check the device connetivity status and don't start sync if not connected.
    device.getRemindersHeaders(function (err, headers) {
      if (err) {
        return dssutil.cbError(qCb, logger, err);
      }
      
      // Get all reminders that aren't synced.
      return platform_reminder.queryBySyncStatus(false, function (err, reminders) {
        if (err) {
          return dssutil.cbError(qCb, logger, err);
        }

        // Sync first to the cloud, then retrieve from the cloud.
        return async.series([_syncToCloud.bind(undefined, reminders, headers),
                             _syncFromCloud.bind(undefined, headers)],
                            qCb);
      });
    });
  },
  function (err) {
    // Mark that a sync is no longer in progress.
    var waiters = syncQueue._fullSyncWaiters;
    syncQueue._fullSyncWaiters = [];
    syncQueue._fullSyncInProgess = false;    
    syncQueue._lastSyncTime = Date.now();

    logger.info('full sync completed with(out) error: ' + err);

    // Execute the callback for the callers waiting on the sync.
    waiters.forEach(function (waiterCb) {
      waiterCb(err);
    });
  });
}

function _syncToCloud(reminders, headers, cb) {
  if (reminders.length === 0) {
    // nothing to sync.
    return cb();
  }
  
  var updatePayload = [];
  var updateReminders = [];
  var deletePayload = [];
  var deleteReminders = [];
  
  // Determine whether the reminder is a local update or local delete.
  for (var i = 0; i < reminders.length; i++) {
    var manifest;
    try {
      manifest = JSON.parse(reminders[i].getExtraData()).manifest;
    }
    catch (e) {
      logger.error('error parsing manifest local extraData: ' + reminders[i].getExtraData());
      continue;
    }

    if (manifest.reminderStatus === types.ReminderStatus.Cancelled) {
      deletePayload.push({ id : manifest.id, lastUpdatedAt : manifest.lastUpdatedAt });
      deleteReminders.push(reminders[i]);
    }
    else {
      if (manifest.reminderStatus === types.ReminderStatus.Snoozed) {
        // N.B.: Currently the server deletes Snoozed reminders and is deprecating
        // the status to be local only.
        // Keep it active, any new updates will overwrite local
        // data, somewhat expected behavior if the user edits.
        manifest.reminderStatus = types.ReminderStatus.Active;
      }

      updatePayload.push(manifest);
      updateReminders.push(reminders[i]);
    }
  }
  
  // Update the cloud with the deleted and created/updated reminders.
  return async.series([_syncDeleted.bind(undefined, deletePayload, headers, deleteReminders),
                       _syncUpdated.bind(undefined, updatePayload, headers, updateReminders)],
                      cb);
}

function _syncUpdated(updatePayload, headers, reminders, cb) {
  if (updatePayload.length === 0) {
    return cb();
  }
  
  logger.info('_syncUpdated payload to cloud: %j', updatePayload);

  var url = agentConfig.getSyncServiceUrl() + REMINDER_CREATE;
  
  // Post the created/updated payload to the cloud, an update is a CREATE reminder with an ID already set.
  return request.post(url, { json: updatePayload, headers: headers }, function (err, response, bodyJson) {
    if (err) {
      return dssutil.cbError(cb, logger, 'err: %s', err);
    }

    if (response.statusCode !== 200 || !bodyJson) {
        return dssutil.cbError(cb, logger, 'status: %s, entity: %j', response.statusCode, bodyJson);
    }

    logger.info('_syncUpdated payload from cloud: %j', bodyJson);
    bodyJson.internalResponses.forEach(function (result, i) {
      if (!result.success) {
        logger.error('itemError: %j', result);
        return;
      }
      
      // Store the sync status and the server ID in the Db.
      reminders[i].setSyncStatus(true);
      reminders[i].setServerId(result.id);
      var extraJson;
      try {
        extraJson = JSON.parse(reminders[i].getExtraData());
      }
      catch (e) {
        logger.error('error parsing manifest local extraData: ' + reminders[i].getExtraData());
        return;
      }
      extraJson.manifest.id = result.id;
      reminders[i].setExtraData(JSON.stringify(extraJson));

      platform_reminder.createOrUpdateReminder(reminders[i]);
    });

    return cb();
  });
}

function _syncDeleted(deletePayload, headers, reminders, cb) {
  if (deletePayload.length === 0) {
    return cb();
  }

  logger.info('_syncDeleted payload to cloud: %j', deletePayload);

  var url = agentConfig.getSyncServiceUrl() + REMINDER_DELETE;
  // Post the deleted payload to the cloud.
  return request.post(url, { json: deletePayload, headers: headers }, function (err, response, bodyJson) {
    if (err) {
        return dssutil.cbError(cb, logger, 'err: %s', err);
    }

    if (response.statusCode !== 200 || !bodyJson) {
       return dssutil.cbError(cb, logger, 'status: %s, entity: %j', response.statusCode, bodyJson);
    }
    
    logger.info('_syncDeleted payload from cloud: %j', bodyJson);
    
    bodyJson.internalResponses.forEach(function (result, i) {
      if (!result.success) {
        logger.error('itemError: %j', result);
        return;
      }
      
      // Delete the reminder locally now that it is synced.
      platform_reminder.deleteByLocalId(reminders[i].getLocalId());
    });

    return cb();
  });
}

function _syncFromCloud(headers, cb) {
  var url = agentConfig.getSyncServiceUrl() + REMINDER_GET;
  
  // The etag marks the state of the DB for the last full sync.  The server will return
  // a 304 if nothing has changed since then.  Otherwise the entire state is returned.
  headers[HALSEY_ETAG_HEADER] = config.getStringConfig(REMINDER_ETAG_KEY);
  logger.info('_syncFromCloud request eTag: ' + headers[HALSEY_ETAG_HEADER]);
  
  request.get(url, { json: true, headers: headers }, function (err, response, bodyJson) {
    if (err) {
        return dssutil.cbError(cb, logger, 'err: %s', err);
    }

    if (response.statusCode !== 200 || !bodyJson) {
      if ((response && response.statusCode) === 304) {
        logger.info('304 NOT MODIFIED');
        return cb();
      }
      else {
        return dssutil.cbError(cb, logger, 'status: %s, entity: %j', err, response.statusCode, bodyJson);
      }
    }
    
    logger.info('_syncFromCloud payload result: %j', bodyJson);
    
    // The server returns a minimal set of the reminders which we then need to request back with the set
    // of IDs we want to download.
    var results = bodyJson.internalResponses || [];
    var payload = [];
    var allIds = [];
    
    async.each(results, function (result, eCb) {
      if (!result.success || !result.manifest) {
        logger.error('itemError: %j', result);
        return eCb();
      }
      
      logger.info('receieved reminder ' + result.id + ' from cloud response.');
      allIds.push(result.id);
      
      // look up the local copy of the reminder and see whether this version is more recently updated.
      return _isReminderUpdateNeeded(result, function (err, needed) {
        // Newer version available.  Request the entire payload.
        if (!err && needed) {
          payload.push(result.id);
        }
        eCb();
      });
    },
    function () {
      var updated = false;
      return async.parallel([
        function (parallelCb) {
          // The server does not indicate in anyway that a reminder was deleted since the last full sync.
          // make sure we didn't miss any and delete them locally if we did.          
          return platform_reminder.queryWhereServerIdNot(allIds, function (err, missedDeletes) {
            if (err) {
              // don't fail the sync just because we couldn't delete leftover reminders.
              return parallelCb();
            }
            
            return async.each(missedDeletes, function (localId, eachCb) {
              logger.info('deleting reminder ' + localId + ' because not in cloud response.');
              updated = true;
              return crud.deleteReminder(localId, null, true, eachCb);
            },
            parallelCb);
          });
        },

        function (parallelCb) {
          if (payload.length === 0) {
            config.setStringConfig(REMINDER_ETAG_KEY, bodyJson.eTag);
            logger.info('_syncFromCloud response eTag: ' + bodyJson.eTag);
            return parallelCb();
          }
          
          // Request all of the needed Ids from the cloud.
          return _createOrUpdateFromCloud(payload, headers, function (err) {
            if (err) {
              return dssutil.cbError(parallelCb, logger, err);
            }
            
            // Log the Etag so the next full sync will get 304 if nothing has changed since this result.
            config.setStringConfig(REMINDER_ETAG_KEY, bodyJson.eTag);
            logger.info('_syncFromCloud response eTag: ' + bodyJson.eTag);
            updated = true;
            return parallelCb();
          });
        }],
      function (err) {
        if (!err && updated) {
          // Notify listeners on the device that an update occurred.
          platform_reminder.notifyReminderUpdate();
        }
        return cb(err);
      });
    });
  });
}

function _isReminderUpdateNeeded(reminderResponse, cb) {
  return platform_reminder.queryByServerId(reminderResponse.id, function (err, reminderDescriptor) {
    if (err) {
      logger.info('reminder %s not on device, requesting updated.', reminderResponse.id);
      return cb(null, true);
    }
    
    var extraJson;
    try {
      extraJson = JSON.parse(reminderDescriptor.getExtraData());
    }
    catch (e) {
      logger.error('error parsing reminder %s json, trying for an update: %s', reminderResponse.id, e);
      return cb(null, true);
    }
    
    if (extraJson === undefined ||
        extraJson.manifest === undefined ||
        extraJson.manifest.lastUpdatedAt === undefined ||
        reminderResponse.manifest === undefined ||
        reminderResponse.manifest.lastUpdatedAt === undefined) {
      logger.error('unexpected reminder json, trying for an update.', extraJson);
      return cb(null, true);
    }
    
    if (new Date(extraJson.manifest.lastUpdatedAt).getTime() < new Date(reminderResponse.manifest.lastUpdatedAt).getTime()) {
      logger.log('newer reminder %s version on server, requesting update', reminderResponse.id);
      return cb(null, true);
    }
    
    logger.log('skipping reminder %s update since local is same or newer.', reminderResponse.id);
    return cb(null, false);
  });
}

function _createOrUpdateFromCloud(payload, headers, cb) {
  logger.info('payload to cloud: %j', payload);
  
  var url = agentConfig.getSyncServiceUrl() + REMINDER_GET;
  // Post the request for a full manifest for each ID in the payload.
  return request.post(url, { json: payload, headers: headers }, function (err, response, bodyJson) {
    if (err) {
        return dssutil.cbError(cb, logger, 'err: %s', err);
    }
    if (response.statusCode !== 200 || !bodyJson) {
      return dssutil.cbError(cb, logger, 'status: %s, entity: %j', response.statusCode, bodyJson);
    }

    logger.info('getFromCloud payload result: %j', bodyJson);
    if (!bodyJson.internalResponses) {
      return dssutil.cbError(cb, logger, 'no responses in request for IDs.');
    }

    // Update the reminder on the device if needed.
    var results = bodyJson.internalResponses;
    return async.each(results, function (result, feCb) {
      if (!result.success) {
        // Log failures but don't cancel the loop.
        return dssutil.cbError(function () { feCb(); }, logger, 'itemError: %j', result);
      }
      
      logger.info('processing reminder: %j', result.manifest);

      // Log failures but don't cancel the loop.
      return crud.createOrUpdateReminder(result.manifest, true, undefined, function () { feCb(); });
    },
    cb);
  });
}

exports.syncReminders = syncReminders;