﻿// Copyright(c) Microsoft Corporation. All rights reserved.

var platform_reminder = require('./platform-reminder.js');
var platform_timer = require('./platform-timer.js');

var types = require('./types.js');
var v1Types = require('./v1types.js');
var trigger = require('./trigger.js');

var logger = require('logger')(__filename + ' reminderSync');
var dssutil = require('dssutil');
var async = require('async');

platform_timer.onTimer(function (timerDescriptor) {
  var alarmName = timerDescriptor.getName();

  logger.info('alarm received: ' + alarmName);

  var match = alarmName.match(/reminder\.([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})$/);

  if (!match) {
    return;
  }

  var reminderId = match[1];
  
  return trigger.triggerReminder(reminderId);
});

function TimeHandler() {
  if (!(this instanceof TimeHandler)) {
    return new TimeHandler();
  }
  
  this.parseFromV1 = function (reminderManifest, v1Obj, cb) {
    var occurrence = new types.ReminderOccurrence();
    
    if (!v1Types.validate(v1Obj) ||
        v1Obj.time === undefined ||
        v1Obj.recurrence === undefined) {
      return false;
    }
    
    // Time indicates the start of the reminder.  Currently we don't support end dates.
    occurrence.startDate = types.getDateIsoString(v1Obj.time);
    occurrence.endDate = '9999-12-31T23:59:59.9999999Z';
    occurrence.occurs = v1Types.ReminderRecurrenceV1toV2Map[v1Obj.recurrence];
    
    // Copy to the V2 manifest.
    reminderManifest.reminderType = types.ReminderType.Time;
    reminderManifest.title = v1Obj.title;
    reminderManifest.description = v1Obj.detail;
    reminderManifest.occurrence = occurrence;
    reminderManifest.reminderStatus = v1Types.ReminderStatusV1toV2Map[v1Obj.status];
    return true;
  };
  
  this.parseToV1 = function parseToV1(v1Reminder, results, reminderManifest) {
    // Convert from the V2 format to the V1 format.  v1Reminder is already set with the common fields.
    
    v1Reminder.time = new Date(reminderManifest.occurrence.startDate).getTime();
    v1Reminder.recurrence = v1Types.ReminderRecurrenceV2toV1Map[reminderManifest.occurrence.occurs];
    
    if (isNaN(v1Reminder.time) ||
        v1Reminder.recurrence === undefined) {
      return false;
    }
    
    results.push({ timeReminder : [v1Reminder] });
    return true;
  };
  
  this.createReminder = function (extraJson, reminderDescriptor, cb) {
    var self = this;

    return async.series([
      // make sure the timer is set on the device.
      function (sCb) {
        // Set the timer on the device if it is Active and we don't already have an alarm.
        if (extraJson.manifest.reminderStatus !== types.ReminderStatus.Active ||
          extraJson.timerId !== undefined) {
          return sCb();
        }
        
        return self._setTimerFromOccurrence(extraJson.manifest.occurrence, reminderDescriptor.getLocalId(), function (err, id) {
          if (!id) {
            return dssutil.cbError(sCb, logger, 'could not set time reminder.');
          }
          
          extraJson.timerId = id;
          return sCb();
        });
      },

      // persist the timer data to the reminders db.
      function (sCb) {
        logger.verbose('createData: %j', extraJson.manifest);
        
        // Use the extraData field to store the relation with the timerId and the full manifest which
        // is opaque data to the db.
        reminderDescriptor.setType(types.ReminderType.Time)
          .setExtraData(JSON.stringify(extraJson));
        
        return platform_reminder.createOrUpdateReminder(reminderDescriptor, function (err) {
          if (err) {
            return dssutil.cbError(sCb, logger, 'could not set reminder');
          }
          
          return sCb(null, reminderDescriptor.getLocalId());
        });
      }
    ],
    // send the reminder ID back to the caller.
    function (err, result) {
      if (err) {
        return dssutil.cbError(cb, logger, 'could not set reminder');
      }

      return cb(null, result[1]);
    });
  };
  
  this.unregisterReminder = function (existingExtraJson, newManifest, cb) {
    if (!existingExtraJson.timerId) {
      return cb();
    }
    
    if (newManifest !== undefined &&
        newManifest.reminderStatus === types.ReminderStatus.Active &&
        existingExtraJson.manifest !== undefined) {
      if (types.areEqualOccurrences(existingExtraJson.manifest.occurrence, newManifest.occurrence)) {
        // The new location matches the existing location ... don't remove them.
        logger.verbose('skipping timer delete because location matches existing %j -- %j', newManifest.occurrence, existingExtraJson.manifest.occurrence);
        return cb();
      }
    }
    
    return platform_timer.cancelPersistentTimer(existingExtraJson.timerId, function (err) {
      if (err) {
        return dssutil.cbError(cb, logger, 'could not cancel persistent timer');
      }

      existingExtraJson.timerId = undefined;
      return cb();
    });
  };
  
  this._setTimerFromOccurrence = function (occurrence, localId, cb) {
    var interval;
    var timerType;
    var now = Date.now();
    var start = new Date(occurrence.startDate).getTime();
    
    switch (occurrence.occurs) {

      case types.ReminderFrequency.Biweekly:
    //TODO: support biweekly natively.
      case types.ReminderFrequency.Weekly:
        timerType = platform_timer.timerType.inexact_repeating;
        // 7 Days interval
        interval = 60 * 60 * 24 * 7 * 1000;
        break;

      case types.ReminderFrequency.TwiceADay:
    //TODO: support twice a day natively.
      case types.ReminderFrequency.FourTimesADay:
    //TODO: support four times a day natively.
      case types.ReminderFrequency.Daily:
        timerType = platform_timer.timerType.inexact_repeating;
        // a Day interval
        interval = 60 * 60 * 24 * 1000;
        break;

      case types.ReminderFrequency.Hourly:
        timerType = platform_timer.timerType.inexact_repeating;
        // an Hour interval
        interval = 60 * 60 * 1000;
        break;

      case types.ReminderFrequency.AsItHappens:
    //TODO: What is AsItHappens?
      case types.ReminderFrequency.Once:
        timerType = platform_timer.timerType.one_time;
        interval = 0;
        break;
      default:
        logger.error('invalid time occurrence specified: ' + occurrence.occurs);
        return null;
    }
    
    // Find the last time the alarm should have fired, or it may be in the future.
    if (start <= now && interval !== 0) {
      start += (interval * (Math.round((now - start) / interval) + 1));
    }
    
    // TODO: use setters instead.
    var timerDescriptor = platform_timer.createTimerDescriptor(timerType,
                                                               'reminder.' + localId,
                                                               start,
                                                               interval,
                                                               '');

    // do not create timers for old reminders with occurrence: once
    if (interval === 0 && start < now) {
      var id = timerDescriptor.getId();
      logger.debug('skipping adding timer for an old time reminder %s', id);
      return cb(null, id);
    }
    
    return platform_timer.createPersistentTimer(timerDescriptor, function (err) {
      if (err) {
        return dssutil.cbError(cb, logger, err);
      }
      
      return cb(null, timerDescriptor.getId());
    });
  };
}

exports.TimeHandler = TimeHandler;