﻿// Copyright(c) Microsoft Corporation. All rights reserved.

var platform_reminder = require('./platform-reminder.js');
var platform_timer = require('./platform-timer.js');

var types = require('./types.js');
var v1Types = require('./v1types.js');

var logger = require('logger')(__filename + ' reminderSync');
var async = require('async');
var dssutil = require('dssutil');
var crudV1 = require('./crudv1.js');

// Default to 10 minutes snooze.
// TODO: Windows supports various times.
var DEFAULT_SNOOZE_TIME_MS = 10 * 60 * 1000;
var SNOOZE_PREFIX = 'reminder.snooze.';

platform_timer.onTimer(function (timerDescriptor) {
  var alarmName = timerDescriptor.getName();
  
  logger.info('alarm received: ' + alarmName);
  
  var match = alarmName.match(/reminder\.snooze\.([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12})$/);
  
  if (!match) {
    return;
  }
  
  var reminderId = match[1];
  logger.info('snoozed reminder received: ' + alarmName);
  
  async.waterfall([
    // first make sure the reminder is still in a snoozed state.
    function (wfCb) {
      return platform_reminder.queryByLocalId(reminderId, function (err, reminderDescriptor) {
        if (err) {
          return dssutil.cbError(wfCb, logger, 'could not find reminder to snooze', err);
        }
        
        if (reminderDescriptor.getStatus() !== types.ReminderStatus.Snoozed) {
          return wfCb('reminder not snoozed.');
        }
        
        return wfCb(null, reminderDescriptor);
      });
    },
    // next switch the reminder to active.
    function (reminderDescriptor, wfCb) {
      reminderDescriptor.setStatus(types.ReminderStatus.Active);
      platform_reminder.createOrUpdateReminder(reminderDescriptor, function (err) {
        if (err) {
          return dssutil.cbError(wfCb, logger, 'could not update snoozed reminder status', err);
        }
        
        return wfCb();
      });
    },
    // next trigger the reminder.
    function (wfCb) {
      logger.info('triggering snoozed reminder: ' + reminderId);
      triggerReminder(reminderId);
      wfCb();
    }]);
});

function snoozeReminder(reminderId, cb) {
  cb = cb || Function();
  
  logger.info('snoozing reminder: ' + reminderId);
  
  // create a one time alarm to trigger snooze.
  var timerDescriptor = platform_timer.createTimerDescriptor(platform_timer.timerType.one_time,
                                                             SNOOZE_PREFIX + reminderId,
                                                             Date.now() + DEFAULT_SNOOZE_TIME_MS,
                                                             0,
                                                             '');

  return platform_timer.createPersistentTimer(timerDescriptor, cb);
}

function triggerReminder(reminderId) {
  return  crudV1.readReminderAsV1(reminderId, function(err, result) {
    if (err) {
      logger.info('skipping reminder trigger, reminder not found: ' + reminderId);
      return;
    }

    var status;

    if (result.timeReminder) {
      status = result.timeReminder[0].status;
    }
    else if (result.businessLocationReminder) {
      status = result.businessLocationReminder[0].status;
    }
    else if (result.locationReminder) {
      status = result.locationReminder[0].status;
    }

    var statusV2 = v1Types.ReminderStatusV1toV2Map[status];
    logger.info('reminder status: ' + statusV2);

    if (statusV2 == types.ReminderStatus.Active || statusV2 == types.ReminderStatus.Snoozed) {
      // TODO: Should we track with the reminder was already triggered?
      // If we prevent recreating the reminder triggers in update it should be enough to not
      // have multiple triggers.
      logger.info('triggering reminder: ' + reminderId);
      var jsonString = JSON.stringify(result);
      platform_reminder.triggerReminder(jsonString);
      return;
    }

    logger.info('skipping reminder trigger, reminder not Active: ' + reminderId);
  });
}

exports.triggerReminder = triggerReminder;
exports.snoozeReminder = snoozeReminder;
