// Copyright(c) Microsoft Corporation. All rights reserved.
var logger = require('logger')(__filename + ' reminderSync');
var util = require('util');

// Definition of Reminder Manifest with defaults for required fields.
// http://bing/wiki/Reminder_service

var DEFAULT_RADIUS_METERS = 200;
var CONTEXT = '@context';
var TYPE = '@type';

var reminderType = {
  Time : 'Time',
  Location : 'Location',
  TimeLocation : 'TimeLocation',
  People: 'People'
};

var reminderStatus = {
  Active : 'Active',
  Inactive : 'Inactive',
  Completed : 'Completed',
  Cancelled : 'Cancelled',
  Snoozed : 'Snoozed'
};

var reminderGeoFenceType = {
  None: 'None',
  Arrive: 'Arrive',
  Depart: 'Depart',
  NearBy: 'NearBy'
};

var reminderFrequency = {
  AsItHappens: 'AsItHappens',
  Weekly: 'Weekly',
  Daily: 'Daily',
  Hourly: 'Hourly',
  TwiceADay: 'TwiceADay',
  FourTimesADay: 'FourTimesADay',
  Biweekly: 'Biweekly',
  Once: 'Once'
};

function ReminderPerson() {
  if (!this instanceof ReminderPerson) {
    return new ReminderPerson();
  }
  
  this[TYPE] = 'Person';
  this.alternateName = 'self';
}

function ReminderDay() {
  if (!this instanceof ReminderDay) {
    return new ReminderDay();
  }
  
  this[TYPE] = 'DayOfWeek';
  this.name = '';
}

function ReminderOccurrence() {
  if (!this instanceof ReminderOccurrence) {
    return new ReminderOccurrence();
  }
  
  this[TYPE] = 'Occurrence';
  this.startDate = '';
  this.endDate = '';
  this.occurs = '';
  this.Days = [];
}

function areEqualOccurrences(occursOne, occursTwo) {
  if (occursOne === undefined && occursTwo === undefined) {
    return true;
  }
  
  if (occursOne === undefined || occursTwo === undefined) {
    return false;
  }
  
  if (occursOne.startDate !== occursTwo.startDate ||
      occursOne.endDate !== occursTwo.endDate ||
      occursOne.occurs !== occursTwo.occurs) {
    return false;
  }
  
  if (occursOne.Days === undefined && occursTwo.Days === undefined) {
    return true;
  }
  
  if (occursOne.Days === undefined || occursTwo.Days === undefined) {
    return false;
  }
  
  if (occursOne.Days.length !== occursTwo.Days.length) {
    return false;
  }
  
  // Asssuming the list is ordered for now.
  if (JSON.stringify(occursOne.Days) !== JSON.stringify(occursTwo.Days)) {
    return false;
  }
  
  return true;
}

function ReminderEntity() {
  if (!this instanceof ReminderEntity) {
    return new ReminderEntity();
  }
  
  this.id = '';
  this.url = '';
  this.type = '';
}

function areEqualEntities(entityOne, entityTwo) {
  if (entityOne === undefined && entityTwo === undefined) {
    return true;
  }
  
  if (entityOne === undefined || entityTwo === undefined) {
    return false;
  }
  
  if (entityOne.id !== entityTwo.id ||
      entityOne.url !== entityTwo.url ||
      entityOne.type !== entityTwo.type) {
    return false;
  }
  
  return true;
}

function ReminderGeo() {
  if (!this instanceof ReminderGeo) {
    return new ReminderGeo();
  }
  
  this[TYPE] = 'GeoCoordinates';
  this.latitude = '';
  this.longitude = '';
}

function areEqualGeos(geoOne, geoTwo) {
  if (geoOne === undefined && geoTwo === undefined) {
    return true;
  }
  
  if (geoOne === undefined || geoTwo === undefined) {
    return false;
  }
  
  if (geoOne.latitude !== geoTwo.latitude ||
      geoOne.longitude !== geoTwo.longitude) {
    return false;
  }
  
  return true;
}

function ReminderAddress() {
  if (!this instanceof ReminderAddress) {
    return new ReminderAddress();
  }
  
  this[TYPE] = 'PostalAddress';
  this.streetAddress = '';
  this.addressLocality = '';
  this.addressRegion = '';
  this.postalCode = '';
  this.addressCountry = '';
  this.addressSingleLine = '';
}

function areEqualAddresses(addressOne, addressTwo) {
  if (addressOne === undefined && addressTwo === undefined) {
    return true;
  }
  
  if (addressOne === undefined || addressTwo === undefined) {
    return false;
  }
  
  if (addressOne.streetAddress !== addressTwo.streetAddress ||
      addressOne.addressLocality !== addressTwo.addressLocality ||
      addressOne.addressRegion !== addressTwo.addressRegion ||
      addressOne.postalCode !== addressTwo.postalCode ||
      addressOne.addressCountry !== addressTwo.addressCountry ||
      addressOne.addressSingleLine !== addressTwo.addressSingleLine) {
    return false;
  }
    
  return true;
}

function ReminderLocation() {
  if (!this instanceof ReminderLocation) {
    return new ReminderLocation();
  }

  this[TYPE] = 'Place';
  this.name = '';
  this.alternateName = '';
  this.address = new exports.ReminderAddress();
  this.geofenceType = '';
  this.geo = undefined;
  this.radius = DEFAULT_RADIUS_METERS;
  this.entity = new exports.ReminderEntity();
}


function areEqualLocations(locationOne, locationTwo) {
  if (locationOne === undefined && locationTwo === undefined) {
    return true;
  }
  
  if (locationOne === undefined || locationTwo === undefined) {
    return false;
  }

  if (locationOne.name !== locationTwo.name ||
      locationOne.alternateName !== locationTwo.alternateName ||
      locationOne.geofenceType !== locationTwo.geofenceType ||
      locationOne.radius !== locationTwo.radius) {
    return false;
  }
  
  if (!areEqualAddresses(locationOne.address, locationTwo.address)) {
    return false;
  }
  
  if (!areEqualGeos(locationOne.geo, locationTwo.geo)) {
    return false;
  }
  
  if (!areEqualEntities(locationOne.entity, locationTwo.entity)) {
    return false;
  }
  
  return true;
}

function ReminderManifest() {
  if (!this instanceof ReminderManifest) {
    return new ReminderManifest();
  }
  
  this[CONTEXT] = 'http://schema.org/reminder';
  this[TYPE] = 'Reminder';
  this.id = undefined;
  this.title = '';
  this.description = '';
  this.reminderType = '';
  this.createdAt = '';
  this.lastUpdatedAt = '';
  this.lastViewedAt = '';
  this.sender = new exports.ReminderPerson();
  this.recipient = [ new exports.ReminderPerson() ];
  this.location = undefined;
  this.occurrence = undefined;
  this.participant = undefined;
  this.reminderStatus = '';
  this.deepLink = undefined;
  this.file = [];
}

// Get the ISO version of the date 'from' or current time if undefined.
function getDateIsoString(from) {
  var date;
  
  if (from) {
    date = new Date(from);
  }
  else {
    date = new Date();
  }

  return date.toISOString();
}

exports.ReminderType = reminderType;
exports.ReminderStatus = reminderStatus;
exports.ReminderGeoFenceType = reminderGeoFenceType;
exports.ReminderFrequency = reminderFrequency;
exports.ReminderPerson = ReminderPerson;
exports.ReminderDay = ReminderDay;
exports.ReminderOccurrence = ReminderOccurrence;
exports.ReminderEntity = ReminderEntity;
exports.ReminderGeo = ReminderGeo;
exports.ReminderAddress = ReminderAddress;
exports.ReminderLocation = ReminderLocation;
exports.ReminderManifest = ReminderManifest;
exports.getDateIsoString = getDateIsoString;
exports.areEqualOccurrences = areEqualOccurrences;
exports.areEqualLocations = areEqualLocations;